#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 19-01-2016
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2016
# license         : LGPLv3
#==============================================================================

import os
import argparse
import shutil
# import numpy as np
# from scipy.sparse import dok_matrix


# def load_sirms_txt(fname):
#     with open(fname) as f:
#         descr_names = f.readline().strip().split("\t")[1:]
#         case_names = []
#         x = []
#         for line in f:
#             tmp = line.strip().split("\t")
#             case_names.append(tmp[0])
#             x.append(tuple(map(float, tmp[1:])))
#     return descr_names, case_names, np.asarray(x)
#
#
# def load_sirms_svm(fname):
#     descr_names = [v.strip() for v in open(os.path.splitext(fname)[0] + '.colnames').readlines()]
#     case_names = [v.strip() for v in open(os.path.splitext(fname)[0] + '.rownames').readlines()]
#     x = dok_matrix((len(case_names), len(descr_names)), dtype=np.float32)
#     with open(fname) as f:
#         for row, line in enumerate(f):
#             tmp = line.strip().split(' ')
#             for v in tmp:
#                 col, value = v.split(':')
#                 x[row, int(col)] = value
#     return descr_names, case_names, x.toarray()
#
#
# def save_sirms_txt(x, descr_names, case_names, fname):
#     print(np.shape(x))
#     print(len(descr_names))
#     print(len(case_names))
#     with open(fname, 'wt') as f:
#         f.write('Compounds' + '\t' + '\t'.join(descr_names) + '\n')
#         for i, row in enumerate(x):
#             f.write(case_names[i] + '\t' + '\t'.join(map(str, row.tolist())) + '\n')
#
#
# def save_sirms_svm(x, descr_names, case_names, fname):
#     open(os.path.splitext(fname)[0] + '.rownames', 'wt').write('\n'.join(case_names))
#     open(os.path.splitext(fname)[0] + '.colnames', 'wt').write('\n'.join(descr_names))
#     with open(fname, 'wt') as f:
#         for row in x:
#             ids = np.nonzero(row)[0].tolist()
#             row = row[row != 0].tolist()
#             line = [str(i) + ':' + str(v) for i, v in zip(ids, row)]
#             f.write(' '.join(line) + '\n')


def main_params(in_fname, out_fname, file_format):

    if file_format == 'txt':

        with open(in_fname) as f_in:
            with open(in_fname + '.tmp' if out_fname == in_fname else out_fname, 'wt') as f_out:
                header = f_in.readline().strip().split('\t')
                ids = ["|HB|I,I,I,I|" not in item for item in header]
                f_out.write('\t'.join([item for item, id in zip(header, ids) if id]) + '\n')
                for line in f_in:
                    f_out.write('\t'.join([item for item, id in zip(line.strip().split('\t'), ids) if id]) + '\n')
        if out_fname == in_fname:
            shutil.move(in_fname + '.tmp', in_fname)

    elif file_format == 'svm':

        colnames = open(os.path.splitext(in_fname)[0] + '.colnames').readlines()
        new_colnames = [item for item in colnames if "|HB|I,I,I,I|" not in item]

        ids = dict()  # {old_num: new_num, ...} as strings
        for i in range(len(colnames)):
            try:
                ids[str(i)] = str(new_colnames.index(colnames[i]))
            except ValueError:
                continue

        with open(in_fname) as f_in:
            with open(in_fname + '.tmp' if out_fname == in_fname else out_fname, 'wt') as f_out:
                for line in f_in:
                    items = [tuple(item.split(':')) for item in line.strip().split(' ')]
                    output = []
                    for item in items:
                        id = ids.get(item[0], None)
                        if id is not None:
                            output.append(id + ':' + item[1])
                    f_out.write(' '.join(output) + '\n')

        if out_fname == in_fname:
            open(os.path.splitext(in_fname)[0] + '.colnames', 'wt').writelines(new_colnames)
            shutil.move(in_fname + '.tmp', in_fname)
        else:
            open(os.path.splitext(out_fname)[0] + '.colnames', 'wt').writelines(new_colnames)
            shutil.copyfile(os.path.splitext(in_fname)[0] + '.rownames', os.path.splitext(out_fname)[0] + '.rownames')


def main():

    parser = argparse.ArgumentParser(description='Filter descriptors: discard HB descriptors with all atoms labeled I.')
    parser.add_argument('-i', '--input', metavar='input_descriptors.txt', required=True,
                        help='input file with descriptors in txt or svm formats.')
    parser.add_argument('-o', '--output', metavar='output_descriptors.txt', default=None,
                        help='output file with descriptors in txt or svm formats. If missing then input file will be '
                             'rewritten through temporary file.')
    parser.add_argument('-f', '--file_format', metavar='txt|svm', default='txt',
                        help='input file format txt or svm. Default: txt.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "input": in_fname = v
        if o == "output": out_fname = v
        if o == "file_format": file_format = v
    if file_format not in ['txt', 'svm']:
        print("Input file format is wrong - %s. Only txt and svm are allowed." % file_format)
        exit()
    if out_fname is None:
        out_fname = in_fname

    main_params(in_fname, out_fname, file_format)


if __name__ == '__main__':
    main()



import re
from itertools import product, permutations, combinations
from collections import defaultdict
from rdkit import Chem

__author__ = 'pavel'

patt_remove_map = re.compile("\[\*\:[0-9]+\]")   # to change CC([*:1])O to CC([*])O
patt_remove_isotope_1 = re.compile("\[[0-9]+([a-z]+)H?[0-9]?\]")   # to change CC([42cH3])[*:1] to CC(c)[*:1]
patt_remove_isotope_2 = re.compile("(\[)[0-9]+([a-z]+)H?[0-9]?(\+\])")   # to change CC([42nH+])[*:1] CC([n+])[*:1]


def get_submol(mol, atom_ids):
    bond_ids = []
    for pair in combinations(atom_ids, 2):
        b = mol.GetBondBetweenAtoms(*pair)
        if b:
            bond_ids.append(b.GetIdx())
    m = Chem.PathToSubmol(mol, bond_ids)
    m.UpdatePropertyCache()
    return m


def __get_context_env(mol, radius):
    """
    INPUT:
        mol - Mol object containing chain(s) of molecular context
        radius - integer, number of bonds to cut context
    OUTPUT:
        Mol containing only atoms within the specified radius from the attachment point(s).
        All explicit Hs will be stripped.
    """
    # mol is context consisting of one or more groups with single attachment point
    bond_ids = set()
    for a in mol.GetAtoms():
        if a.GetSymbol() == "*":
            i = radius
            b = Chem.FindAtomEnvironmentOfRadiusN(mol, i, a.GetIdx())
            while not b and i > 0:
                i -= 1
                b = Chem.FindAtomEnvironmentOfRadiusN(mol, i, a.GetIdx())
            bond_ids.update(b)
    m = Chem.PathToSubmol(mol, list(bond_ids))
    # remove Hs, otherwise terminal atoms will produce smiles with H ([CH2]C[*:1])
    for a in m.GetAtoms():
        a.SetNumExplicitHs(0)
    return m


def __replace_att(mol, repl_dict):
    for a in mol.GetAtoms():
        map_num = a.GetAtomMapNum()
        if map_num in repl_dict:
            a.SetAtomMapNum(repl_dict[map_num])


def __get_maps_and_ranks(env, keep_stereo=False):
    """
    Return the list of attachment point map numbers and
    the list of canonical SMILES without mapped attachment points (ranks)
    """
    tmp_mol = Chem.Mol(env)
    maps = []
    ranks = []
    for comp in Chem.GetMolFrags(tmp_mol, asMols=True, sanitizeFrags=False):
        for a in comp.GetAtoms():
            atom_num = a.GetAtomMapNum()
            if atom_num:
                maps.append(atom_num)
                a.SetAtomMapNum(0)
                break
        ranks.append(Chem.MolToSmiles(comp, isomericSmiles=keep_stereo))
    return maps, ranks


def __standardize_att_by_env(env, core, keep_stereo=False):
    """
    Set attachment point numbers in core and context according to canonical ranks of attachment points in context
    Ties are broken
    Makes changes in place
    """
    maps, ranks = __get_maps_and_ranks(env, keep_stereo)
    new_att = {m: i+1 for i, (r, m) in enumerate(sorted(zip(ranks, maps)))}
    __replace_att(core, new_att)
    __replace_att(env, new_att)


def __get_att_permutations(env):
    """
    Return possible permutations of attachment point map numbers as a tuple of dicts,
    where each dict: key - old number, value - new number
    """
    maps, ranks = __get_maps_and_ranks(env)

    d = defaultdict(list)
    for rank, att in zip(ranks, maps):
        d[rank].append(att)

    c = []
    for v in d.values():
        c.append([dict(zip(v, x)) for x in permutations(v, len(v))])

    return tuple(__merge_dicts(*item) for item in product(*c))


def __permute_att(mol, d):
    new_mol = Chem.Mol(mol)
    for a in new_mol.GetAtoms():
        i = a.GetAtomMapNum()
        if i in d:
            a.SetAtomMapNum(d[i])
    return new_mol


def __merge_dicts(*dicts):
    res = dicts[0].copy()
    for item in dicts[1:]:
        res.update(item)
    return res


def __standardize_smiles_with_att_points(mol, keep_stereo=False):
    """
    to avoid different order of atoms in SMILES with different map number of attachment points

    smi = ["ClC1=C([*:1])C(=S)C([*:2])=C([*:3])N1",
           "ClC1=C([*:1])C(=S)C([*:3])=C([*:2])N1",
           "ClC1=C([*:2])C(=S)C([*:1])=C([*:3])N1",
           "ClC1=C([*:2])C(=S)C([*:3])=C([*:1])N1",
           "ClC1=C([*:3])C(=S)C([*:1])=C([*:2])N1",
           "ClC1=C([*:3])C(=S)C([*:2])=C([*:1])N1"]

    these will produce different output with RDKit MolToSmiles():
        S=c1c([*:1])c(Cl)[nH]c([*:3])c1[*:2]
        S=c1c([*:1])c(Cl)[nH]c([*:2])c1[*:3]
        S=c1c([*:1])c([*:3])[nH]c(Cl)c1[*:2]
        S=c1c([*:2])c(Cl)[nH]c([*:1])c1[*:3]
        S=c1c([*:1])c([*:2])[nH]c(Cl)c1[*:3]
        S=c1c([*:2])c([*:1])[nH]c(Cl)c1[*:3]

    output of this function
        S=c1c([*:2])c([*:3])[nH]c(Br)c1[*:1]
        S=c1c([*:3])c([*:2])[nH]c(Br)c1[*:1]
        S=c1c([*:1])c([*:3])[nH]c(Br)c1[*:2]
        S=c1c([*:3])c([*:1])[nH]c(Br)c1[*:2]
        S=c1c([*:1])c([*:2])[nH]c(Br)c1[*:3]
        S=c1c([*:2])c([*:1])[nH]c(Br)c1[*:3]

    https://sourceforge.net/p/rdkit/mailman/message/35862258/
    """

    # update property cache if needed
    if mol.NeedsUpdatePropertyCache():
        mol.UpdatePropertyCache()

    # store original maps and remove map numbers from mol
    backup_atom_map = "backupAtomMap"
    for a in mol.GetAtoms():
        atom_map = a.GetAtomMapNum()
        if atom_map:
            a.SetIntProp(backup_atom_map, atom_map)
            a.SetAtomMapNum(0)

    # get canonical ranks for atoms for a mol without maps
    atoms = list(zip(list(Chem.CanonicalRankAtoms(mol)), [a.GetIdx() for a in mol.GetAtoms()]))
    atoms.sort()

    # set new atom maps based on canonical order
    rep = {}
    atom_map = 1
    for pos, atom_idx in atoms:
        a = mol.GetAtomWithIdx(atom_idx)
        if a.HasProp(backup_atom_map):
            a.SetAtomMapNum(atom_map)
            rep["[*:%i]" % atom_map] = "[*:%i]" % a.GetIntProp(backup_atom_map)
            atom_map += 1

    # get SMILES and relabel with original map numbers
    s = Chem.MolToSmiles(mol, isomericSmiles=keep_stereo)
    rep = dict((re.escape(k), v) for k, v in rep.items())
    patt = re.compile("|".join(rep.keys()))
    s = patt.sub(lambda m: rep[re.escape(m.group(0))], s)

    return s


def __get_std_env_smiles(env):
    # env: Mol
    #
    # The issue is different SMILES if atom ids (order) is different. RDKit ignores atom types during canonicalization.
    # m = RWMol()
    #
    # for i in range(3):
    #     a = Atom(6)
    #     m.AddAtom(a)
    # a = Atom(0)
    # m.AddAtom(a)
    #
    # m.GetAtomWithIdx(0).SetIsAromatic(True)  # set atom 0 as aromatic
    # m.GetAtomWithIdx(3).SetAtomMapNum(1)
    #
    # m.AddBond(0, 1, Chem.rdchem.BondType.SINGLE)
    # m.AddBond(1, 2, Chem.rdchem.BondType.SINGLE)
    # m.AddBond(1, 3, Chem.rdchem.BondType.SINGLE)
    #
    # Chem.MolToSmiles(m)
    #
    # OUTPUT: 'cC(C)[*:1]'
    #
    # # 2 ===========
    #
    # m2 = RWMol()
    #
    # for i in range(3):
    #     a = Atom(6)
    #     m2.AddAtom(a)
    # a = Atom(0)
    # m2.AddAtom(a)
    #
    # m2.GetAtomWithIdx(2).SetIsAromatic(True)  # set atom 2 as aromatic
    # m2.GetAtomWithIdx(3).SetAtomMapNum(1)
    #
    # m2.AddBond(0, 1, Chem.rdchem.BondType.SINGLE)
    # m2.AddBond(1, 2, Chem.rdchem.BondType.SINGLE)
    # m2.AddBond(1, 3, Chem.rdchem.BondType.SINGLE)
    #
    # Chem.MolToSmiles(m2)
    #
    # OUTPUT: 'CC(c)[*:1]'
    #
    #
    # Solution: set isotop for all aromatic atoms and then replace in output SMILES
    # Improvement 2: since the canon algorithm depends on number of hydrogens attached to an atom set no implicit Hs

    for a in env.GetAtoms():
        if a.GetIsAromatic():
            a.SetIsotope(42)
        a.SetNoImplicit(True)
    s = Chem.MolToSmiles(env, isomericSmiles=True)
    return patt_remove_isotope_2.sub("\\1\\2\\3", patt_remove_isotope_1.sub('\\1', s))


def get_std_context_core_permutations(context, core, radius, keep_stereo):
    """
    INPUT:
        context - Mol or SMILES containing full chain(s) of a context with labeled attachment point(s),
                  if context is absent (e.g.for radius 0) specify empty string or empty Mol
        core    - Mol or SMILES of a core fragment with labeled attachment point(s)
        keep_stereo - boolean to keep stereo information in output
        radius  - integer (0, 1, 2, etc), number of bonds to cut context
    OUTPUT:
        SMILES of a context environment of a specified radius,
        list of SMILES of a core fragment with possible permutations of attachment point numbers
        env_smi, (core_smi_1, core_smi_2, ...)

        env_smi will not contain any Hs

        for radius 0 attachment point numbers will be stripped, but the string will correspond to core SMILES with
        radius > 0 if remove all map numbers from SMILES

    Output SMILES are standardized
    """

    if isinstance(context, str):
        context = Chem.MolFromSmiles(context)
    if isinstance(core, str):
        core = Chem.MolFromSmiles(core)

    # remove Hs from context and core
    if context:  # context cannot be H (no check needed), if so the user will obtain meaningless output
        context = Chem.RemoveHs(context)
    if core and Chem.MolToSmiles(core) != '[H][*:1]':
        core = Chem.RemoveHs(core)

    if radius == 0 and core:

        if not keep_stereo:
            Chem.RemoveStereochemistry(core)

        s = __standardize_smiles_with_att_points(core, keep_stereo)
        s = patt_remove_map.sub("[*]", s)

        return '', (s, )

    if core and context:

        att_num = len(Chem.GetMolFrags(context))

        if not keep_stereo:
            Chem.RemoveStereochemistry(context)
            Chem.RemoveStereochemistry(core)

        env = __get_context_env(context, radius)   # cut context to radius
        __standardize_att_by_env(env, core, keep_stereo)
        env_smi = __get_std_env_smiles(env)

        if att_num == 1:

            return env_smi, (__standardize_smiles_with_att_points(core, keep_stereo), )

        else:

            res = []
            p = __get_att_permutations(env)

            # permute attachment point numbering only in core,
            # since permutations in env will give the same canonical smiles
            if len(p) > 1:
                for d in p:
                    c = __permute_att(core, d)
                    res.append(c)
            else:
                res.append(core)

            # get distinct standardized SMILES
            d = tuple(set(__standardize_smiles_with_att_points(m, keep_stereo) for m in res))

            return env_smi, d

    return None, None


def get_canon_context_core(context, core, radius, keep_stereo=False):
    # context and core are Mols or SMILES
    # returns SMILES by default
    res = get_std_context_core_permutations(context, core, radius, keep_stereo)
    if res:
        env, cores = res
        return env, sorted(cores)[0]
    else:
        return None, None



#!/usr/bin/env python
# author          : Pavel
# date            : 04.04.17
# version         : 0.1
# python_version  : 3
# copyright       : Pavel 2017
# license         : LGPLv3
#==============================================================================

import os
import sys
import argparse
from datetime import datetime
from rdkit import Chem
from rdkit.Chem.Scaffolds import MurckoScaffold
from .find_frags_auto_rdkit import replace_no2


def add_atoms(atom, remove_atoms):
    for a in atom.iterateNeighbors():
        d = set([ai.index() for ai in a.iterateNeighbors()]).difference(remove_atoms)
        if len(d) == 1:
            remove_atoms.append(a.index())
            add_atoms(a, remove_atoms)


def main_params(in_sdf, out_txt, verbose, error_fname):

    with open(out_txt, "wt") as f:

        for i, mol in enumerate(Chem.SDMolSupplier(in_sdf, sanitize=False, removeHs=False)):

            if mol:

                # modify representation of NO2 groups to charged version
                mol = replace_no2(mol)
                err = Chem.SanitizeMol(mol, catchErrors=True)
                if err != 0:
                    with open(error_fname, 'at') as f_err:
                        f_err.write('%s\t%s\t%s\n' % (datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                                      os.path.basename(__file__),
                                                      'Molecule %s failed to sanitize due to: ' % mol.GetProp("_Name") + str(err)))

                s = MurckoScaffold.GetScaffoldForMol(mol)
                ids = mol.GetSubstructMatches(s)
                if len(ids) == 1:
                    if len(ids[0]) > 0:   # to skip mol without scaffold
                        ids = [j + 1 for j in ids[0]]
                        f.write(mol.GetProp("_Name") + "\t" + Chem.MolToSmiles(s) + "\t" + "\t".join(map(str, ids)) + "\n")

                else:
                    sys.stderr.write('More than one scaffold match was found. Molecule: %s, scaffold: %s' %
                                     (Chem.MolToSmiles(mol), Chem.MolToSmiles(s)))
                    sys.stderr.flush()

            else:
                with open(error_fname, 'at') as f_err:
                    f_err.write('%s\t%s\tcould not read molecule number %i from file\n' %
                                (datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                 os.path.basename(__file__), i + 1))

            if verbose and (i + 1) % 100 == 0:
                print('%i molecules passed' % (i + 1))


def main():

    parser = argparse.ArgumentParser(description='Find indices of Murcko framework atoms in molecules from input sdf-file.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles.')
    parser.add_argument('-o', '--out', metavar='output.txt', required=True,
                        help='output text file; each line contains tab-separated fields: name of a molecule, '
                             'name of a fragment and list of corresponding atom numbers.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='show progress on the screen.')
    parser.add_argument('-e', '--error_file', metavar='log_file_name.txt', default="rdkit_errors.txt",
                        help='save names of molecules which cause error to a text log file. Default file name '
                             'rdkit_errors.txt.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_sdf = v
        if o == "out": out_txt = v
        if o == "verbose": verbose = v
        if o == "error_file": error_fname = v

    main_params(in_sdf, out_txt, verbose, error_fname)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# #==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import os
import re
import ast
import glob
import shutil
import tkinter as tk

from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox
from tkinter import font as tkfont
from multiprocessing import cpu_count
from subprocess import call
from collections import OrderedDict

from . import sdf_field2title
from . import calc_atomic_properties_chemaxon
from . import model
from . import find_frags_rdkit as find_frags
# import find_rings_indigo as find_rings
from . import find_murcko_rdkit as find_murcko
from . import find_frags_auto_rdkit as find_frags_auto
from . import calc_frag_contrib
from . import plot_contributions
from . import extractsdf
from . import filter_descriptors
from . import predict
from . import descriptors

from sirms import sirms


def get_script_path():
    return os.path.abspath(os.path.dirname(__file__))


def quote_str(s):
    return '"%s"' % s


class StatWindow(tk.Toplevel):

# it's magic
# ideas were taken from
# http://stackoverflow.com/questions/13205727/using-ttk-treeview-how-can-i-get-horizontal-scrolling-to-work-as-expected

    def close_window(self):
        self.destroy()

    def __read_stat_file(self, stat_fname, keep_only_important_details=True):

        d = dict()
        with open(stat_fname) as f:
            d['model_type'] = f.readline().strip()
            d['header'] = f.readline().strip().split('\t')
            d['lines'] = []
            for line in f:
                d['lines'].append(line.strip().split('\t'))

        if keep_only_important_details:
            for i, line in enumerate(d['lines']):
                if line[1] == 'gbm' or line[1].startswith('gbm_'):
                    names = ['subsample', 'learning_rate', 'n_estimators', 'max_features', 'max_depth']
                elif line[1] == 'rf' or line[1].startswith('rf_'):
                    names = ['n_estimators', 'max_features']
                elif line[1] == 'svm' or line[1].startswith('svm_'):
                    names = ['kernel', 'C', 'gamma']
                elif line[1] == 'knn' or line[1].startswith('knn_'):
                    names = ['n_neighbors']
                elif line[1] == 'pls' or line[1].startswith('pls_'):
                    names = ['n_components']
                else:
                    "Oops! That was a stupid error."
                if len(line) == len(d['header']):
                    tmp = ast.literal_eval('{' + line[-1] + '}')
                    d['lines'][i][-1] = '; '.join([k + ' = ' + str(v) for k, v in tmp.items() if k in names])

        return d

    def fill_data(self, stat_fname):

        d = self.__read_stat_file(stat_fname)

        self.title(d['model_type'] + ' models statistics')

        self.tree.configure(columns=d['header'], show="headings")
        for col in d['header']:
            self.tree.heading(col, text=col)
            self.tree.column(col, width=tkfont.Font().measure(col), stretch=False)
        self.tree.column(d['header'][-1], stretch=True)

        for item in d['lines']:
            self.tree.insert('', 'end', values=item)
            for ix, val in enumerate(item):
                col_w = tkfont.Font().measure(val)
                if self.tree.column(d['header'][ix], width=None) < col_w:
                    self.tree.column(d['header'][ix], width=col_w)

        # resize window
        self.configure(height=len(d['lines']) * 40 + 60)

    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent, width=600)

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.treeFrame = ttk.Frame(self)
        self.treeFrame.grid(row=0, column=0, padx=5, pady=5, sticky='nsew')
        self.treeFrame.grid_columnconfigure(0, weight=1)
        self.treeFrame.rowconfigure(0, weight=1)

        def autoscrollv(sbar, first, last):
            first, last = float(first), float(last)
            if first <= 0 and last >= 1:
                sbar.grid_remove()
            else:
                sbar.grid()
            sbar.set(first, last)

        def autoscrollh(sbar, first, last):
            first, last = float(first), float(last)
            if first <= 0 and last >= 1:
                sbar.grid_remove()
            else:
                sbar.grid()
            sbar.set(first, last)

        vsb = ttk.Scrollbar(self.treeFrame, orient="vertical")
        hsb = ttk.Scrollbar(self.treeFrame, orient="horizontal")

        self.tree = ttk.Treeview(self.treeFrame, height=10,
                                 yscrollcommand=lambda f, l: autoscrollv(vsb, f, l),
                                 xscrollcommand=lambda f, l: autoscrollh(hsb, f, l))
        self.tree.column("#0", minwidth=400, stretch=True)

        vsb['command'] = self.tree.yview
        hsb['command'] = self.tree.xview

        self.tree.grid(row=0, column=0, sticky='nsew')
        vsb.grid(column=1, row=0, sticky='ns')
        hsb.grid(column=0, row=1, sticky='ew')

        ttk.Button(self.treeFrame, text='OK', command=self.close_window).grid(column=0, row=2)

        self.grid_propagate(False)
        self.configure(width=600, height=150)


class SpinboxCPUFrame(ttk.Labelframe):

    def __init__(self, parent, column, row, name="Number of cores to use"):

        ttk.Labelframe.__init__(self, parent, text=name)
        self.grid(column=column, row=row, columnspan=10, sticky=(tk.W, tk.E), padx=5, pady=5)

        self._value = tk.StringVar(value=str(cpu_count()-1))
        tk.Spinbox(self, from_=1, to=cpu_count(), textvariable=self._value, width=5).grid(column=0, row=2, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))

    def get_value(self):
        return int(self._value.get())


class ModelsFrame(ttk.Labelframe):

    def __init__(self, parent, name, column, row, full_version=True):

        ttk.Labelframe.__init__(self, parent, text=name)
        self.grid(column=column, row=row, columnspan=10, sticky=(tk.W, tk.E), padx=5, pady=5)

        self._full_version = full_version

        self.model_type = tk.StringVar(value='reg')

        self.model_rf = tk.BooleanVar(value=True)
        self.model_svm = tk.BooleanVar(value=True)
        self.model_gbm = tk.BooleanVar(value=True)
        self.model_pls = tk.BooleanVar(value=True)
        self.model_knn = tk.BooleanVar(value=False)

        self.model_rf_class = tk.BooleanVar(value=True)
        self.model_svm_class = tk.BooleanVar(value=True)
        self.model_gbm_class = tk.BooleanVar(value=True)
        self.model_knn_class = tk.BooleanVar(value=False)

        self.model_imbalanced = tk.BooleanVar(value=False)

        ttk.Radiobutton(self, text='Regression (RF, GBM, SVM, PLS)', name='model_type_reg', value='reg', variable=self.model_type).grid(column=0, row=0, sticky=(tk.W), padx=5, pady=5)
        ttk.Radiobutton(self, text='Binary classification (0-1) (RF, GBM, SVM)', name='model_type_class', value='class', variable=self.model_type).grid(column=1, row=0, sticky=(tk.W), padx=5, pady=5)

        if self._full_version:

            ttk.Checkbutton(self, variable=self.model_rf, name='chk_rf', text='Random Forest (RF)', command=self.reg_model_checked).grid(column=0, row=1, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_svm, name='chk_svr', text='Support vector regression (SVR)', command=self.reg_model_checked).grid(column=0, row=2, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_gbm, name='chk_gbm', text='Gradient boosting regression (GBR)', command=self.reg_model_checked).grid(column=0, row=3, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_pls, name='chk_pls', text='Partial least squares (PLS)', command=self.reg_model_checked).grid(column=0, row=4, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_knn, name='chk_knn', text='k-Nearest neighbors (kNN)', command=self.reg_model_checked).grid(column=0, row=5, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))

            ttk.Checkbutton(self, variable=self.model_rf_class, name='chk_rf_class', text='Random Forest (RF)', command=self.class_model_checked).grid(column=1, row=1, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_svm_class, name='chk_svr_class', text='Support vector classification (SVC)', command=self.class_model_checked).grid(column=1, row=2, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_gbm_class, name='chk_gbm_class', text='Gradient boosting classification (GBC)', command=self.class_model_checked).grid(column=1, row=3, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))
            ttk.Checkbutton(self, variable=self.model_knn_class, name='chk_knn_class', text='k-Nearest neighbors (kNN)', command=self.class_model_checked).grid(column=1, row=4, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))

        ttk.Checkbutton(self, variable=self.model_imbalanced, name='chk_imbalanced', text='Use multiple undersampling (for imbalanced data sets)').grid(column=1, row=10, sticky=(tk.W, tk.E), padx=5, pady=(1, 0))

        self.columnconfigure(0, pad=80)

    def class_model_checked(self):
        self.model_type.set(value='class')
        # self.children['model_type_class'].set(value=True)

    def reg_model_checked(self):
        self.model_type.set(value='reg')
        # self.children['model_type_reg'].set(value=True)

    def get_selected_models(self):

        output = []
        if self.model_type.get() == 'reg':
            if self._full_version:
                if self.model_gbm.get(): output.append('gbm')
                if self.model_rf.get(): output.append('rf')
                if self.model_svm.get(): output.append('svm')
                if self.model_pls.get(): output.append('pls')
                if self.model_knn.get(): output.append('knn')
            else:
                output = ['gbm', 'rf', 'svm', 'pls']
        if self.model_type.get() == 'class':
            if self._full_version:
                if self.model_gbm_class.get(): output.append('gbm')
                if self.model_rf_class.get(): output.append('rf')
                if self.model_svm_class.get(): output.append('svm')
                if self.model_knn_class.get(): output.append('knn')
            else:
                output = ['gbm', 'rf', 'svm']

        return output

    def get_selected_models_with_type(self):
        models = self.get_selected_models()
        return [m + '_' + self.model_type.get() for m in models]

    def get_imbalanced(self):
        return self.model_imbalanced.get()


class Tab_1(ttk.Frame):

    def __select_chemaxon_dir(self):
        self.chemaxon_dir.set(filedialog.askdirectory())

    def __select_sdf_path(self):
        self.sdf_path.set(filedialog.askopenfilename(filetypes=[('SDF files', '*.sdf')]))

    def __select_property_file_path(self):
        self.property_file_path.set(filedialog.askopenfilename(filetypes=[('Property text file', '*.txt')],
                                                               initialdir=os.path.dirname(self.sdf_path.get())))

    def __remove_forbidden_y(self, filename):
        # remove non-numeric items from y file (with compounds properties)

        def is_number(s):
            try:
                float(s)
                return True
            except ValueError:
                return False

        lines = open(filename).readlines()
        # add header to the output
        output = [lines[0]]
        for line in lines[1:]:
            if is_number(line.strip().split('\t')[1]):
                output.append(line)
        open(filename, 'wt').writelines(output)

    def create_mol_titles(self, in_fname, out_fname):
        if self.compound_names.get() == 'title':
            shutil.copyfile(in_fname, out_fname)
        elif self.compound_names.get() == 'gen':
            sdf_field2title.main_params(in_fname, None, out_fname)
        elif self.compound_names.get() == 'field':
            sdf_field2title.main_params(in_fname, self.sdf_id_field_name.get().strip(), out_fname)
        else:
            print("error occurred which should never be possible")

    def __build_models(self):

        if self.sdf_path.get() == '':
            messagebox.showerror('ERROR!', 'Specify sdf filename and path.')
            return

        output_sdf = self.sdf_path.get().rsplit(".")[0] + '_std_lbl.sdf'

        # if x.txt does not exist then do all standardization and descriptors calculation
        if not os.path.isfile(os.path.join(os.path.dirname(self.sdf_path.get()), 'x.txt')):

            if os.path.isfile(output_sdf):
                os.remove(output_sdf)

            # standardization and property labeling
            if self.descriptors_choice.get() == 'with_chemaxon':

                # standardize
                # print('Standardization is in progress...')
                # copy xml-rules if the file is absent in the sdf folder
                # shutil.copyfile(os.path.join(get_script_path(), 'std_rules.xml'),
                #                 os.path.join(os.path.dirname(self.sdf_path.get()), 'std_rules.xml'))

                # std_sdf_tmp = self.sdf_path.get() + '.std.sdf'
                # run_params = ['standardize',
                #               '-c',
                #               quote_str(os.path.join(os.path.dirname(self.sdf_path.get()), 'std_rules.xml')),
                #               quote_str(self.sdf_path.get()),
                #               '-f',
                #               'sdf',
                #               '-o',
                #               quote_str(std_sdf_tmp)]
                # call(' '.join(run_params), shell=True)

                # calc atomic properties with Chemaxon
                print('Atomic properties calculation is in progress...')
                prop_sdf_tmp = self.sdf_path.get() + '.prop.sdf'
                calc_atomic_properties_chemaxon.main_params(in_fname=self.sdf_path.get(),
                                                            out_fname=prop_sdf_tmp,
                                                            prop=['charge', 'logp', 'acc', 'don', 'refractivity'],
                                                            pH=None,
                                                            cxcalc_path='cxcalc')

                self.create_mol_titles(prop_sdf_tmp, output_sdf)

                # os.remove(std_sdf_tmp)
                os.remove(prop_sdf_tmp)

            else:

                self.create_mol_titles(self.sdf_path.get(), output_sdf)

            print("Descriptors calculation started. Please wait it can take some time")
            x_fname = os.path.join(os.path.dirname(output_sdf), 'x.txt')

            # SiRMS
            if self.descriptors_choice.get() in ['with_chemaxon', 'no_chemaxon']:

                # copy setup.txt to folder with sdf file
                shutil.copyfile(os.path.join(get_script_path(), 'setup.txt'),
                                os.path.join(os.path.dirname(self.sdf_path.get()), 'setup.txt'))

                # calc sirms descriptors
                if self.descriptors_choice.get() == 'with_chemaxon':
                    atom_diff = ['CHARGE', 'LOGP', 'HB', 'REFRACTIVITY']
                else:
                    atom_diff = ['elm']

                sirms.main_params(in_fname=output_sdf,
                                  out_fname=x_fname,
                                  opt_diff=atom_diff,
                                  min_num_atoms=2,
                                  max_num_atoms=4,
                                  min_num_components=1,
                                  max_num_components=2,
                                  min_num_mix_components=2,
                                  max_num_mix_components=2,
                                  mix_fname=None,
                                  descriptors_transformation='num',
                                  mix_type='abs',
                                  opt_mix_ordered=False,
                                  opt_verbose=False,
                                  opt_noH=True,
                                  frag_fname=None,
                                  per_atom_fragments=False,
                                  self_association_mix=False,
                                  reaction_diff=False,
                                  quasimix=False,
                                  id_field_name=None,
                                  output_format='svm',
                                  ncores=min(10, max(self.sb_cpu_count.get_value(), 1)))

                # filter sirms descriptors
                filter_descriptors.main_params(in_fname=x_fname,
                                               out_fname=x_fname,
                                               file_format='svm')

            # RDKit
            else:
                descriptors.main_params(in_fname=output_sdf,
                                        out_fname=x_fname,
                                        output_format='svm',
                                        get_fp=self.descriptors_choice.get(),
                                        opt_verbose=False,
                                        opt_noH=True,
                                        frag_fname=None,
                                        per_atom_fragments=False,
                                        id_field_name=None)

            print("Descriptors calculation finished")

        else:
            x_fname = os.path.join(os.path.dirname(self.sdf_path.get()), 'x.txt')
            self.property_file_path.set(os.path.join(os.path.dirname(self.sdf_path.get()),
                                                     self.property_field_name.get().strip() + '.txt'))

        # extract property, check for numeric values and save to separate file
        if self.property_field_name.get() != '':
            property_filename = os.path.join(os.path.dirname(self.sdf_path.get()),
                                             self.property_field_name.get().strip() + '.txt')
            extractsdf.main_params(in_fname=os.path.splitext(self.sdf_path.get())[0] + '_std_lbl.sdf',
                                   out_fname=property_filename,
                                   title=True,
                                   field_names=[self.property_field_name.get().strip()],
                                   all_fields=False,
                                   skip_value='NA')
            self.__remove_forbidden_y(property_filename)
            self.property_file_path.set(property_filename)

        # remove models from output dir which conflict with newly build models
        models_dir = os.path.join(os.path.dirname(x_fname), self.property_field_name.get().strip(), "models")
        if os.path.isdir(models_dir):
            file_names = glob.glob1(models_dir, '*.pkl')
            for file_name in file_names:
                if file_name.rsplit('.', 1)[0] in self.models_frame.get_selected_models():
                    os.remove(os.path.join(models_dir, file_name))

        # build models
        model.main_params(x_fname=x_fname,
                          y_fname=self.property_file_path.get(),
                          model_names=self.models_frame.get_selected_models(),
                          models_dir=models_dir,
                          ncores=self.sb_cpu_count.get_value(),
                          model_type=self.models_frame.model_type.get(),
                          verbose=1,
                          cv_predictions=True,
                          input_format='svm',
                          imbalanced=self.models_frame.get_imbalanced())

        # update list of models to plot
        self.master.children['tab_3']._show_models_list()

    def __show_models_stat(self):
        win = StatWindow(self)
        win.fill_data(os.path.join(os.path.dirname(self.sdf_path.get()),
                                   self.property_field_name.get().strip(),
                                   'models', 'models_stat.txt'))

    def __add_sdf_path_to_history(self, sdf_path):
        if sdf_path is None or sdf_path == "":
            return None
        max_lines = 10
        new_line = sdf_path + '\n'
        hist = os.path.join(get_script_path(), 'history.txt')
        if os.path.isfile(hist):
            lines = open(hist).readlines()
            # remove newly added line if it is present in lines
            if new_line in lines:
                lines.remove(new_line)
            # keep only allowed number of lines - 1
            if len(lines) >= max_lines:
                lines = lines[-(max_lines-1):]
            lines.append(new_line)
            open(hist, 'wt').writelines(lines)
        else:
            open(hist, 'wt').write(new_line)

    def __read_sdf_history(self):
        file_name = os.path.join(get_script_path(), 'history.txt')
        if os.path.isfile(file_name):
            lines = open(file_name).readlines()
            lines = [line.strip() for line in lines]
            lines.reverse()
            return lines

    def __sdf_path_changed(self, varname, elementname, mode):
        field_names = self.__read_sdf_field_names(self.sdf_path.get())
        if field_names:
            self.children['sdf_label_frame'].children['property_field_name'].configure(values=field_names)
            self.children['sdf_label_frame'].children['property_field_name'].set(value=field_names[0])
            self.children['optional_label_frame'].children['inner_frame'].children['sdf_id_field_name'].configure(values=field_names)
            self.children['optional_label_frame'].children['inner_frame'].children['sdf_id_field_name'].set(value=field_names[0])
        self.__add_sdf_path_to_history(self.sdf_path.get())
        self.children['sdf_label_frame'].children['sdf_path_combobox'].configure(values=self.__read_sdf_history())
        # update list of models to plot
        self.master.children['tab_3']._show_models_list()

    def __prop_field_name_changed(self, varname, elementname, mode):
        # update list of models to plot
        self.master.children['tab_3']._show_models_list()

    def __read_sdf_field_names(self, fname):
        p = re.compile('> {1,2}<(.*)>( +\([0-9]+\))?')
        field_names = set()
        if not os.path.isfile(fname):
            messagebox.showerror('ERROR!', "Specified file name doesn't exist.")
            return field_names
        with open(fname) as f:
            for line in f:
                line = line.rstrip()
                if p.fullmatch(line):
                    field_names.add(p.sub('\\1', line))
        return sorted(field_names)

    def __set_compound_names_choice_event(self, event):
        self.compound_names.set(value='field')

    def __descriptors_choice_changed(self, varname, elementname, mode):
        contr_names = ['contr_overall', 'contr_charge', 'contr_logp', 'contr_hb', 'contr_ref']
        if self.descriptors_choice.get() == 'with_chemaxon':
            states = dict(zip(contr_names, ['normal'] * len(contr_names)))
        else:
            states = dict(zip(contr_names, ['normal', 'disabled', 'disabled', 'disabled', 'disabled']))
            # uncheck all except overall
            self.master.children['tab_3'].contr_charge.set(False)
            self.master.children['tab_3'].contr_logp.set(False)
            self.master.children['tab_3'].contr_hb.set(False)
            self.master.children['tab_3'].contr_ref.set(False)

        for contr_name in contr_names:
            self.master.children['tab_3'].children['contributions'].children[contr_name].config(state=states[contr_name])

    def __init__(self, parent, tab_name):

        ttk.Frame.__init__(self, parent, name='tab_1')

        self.descriptors_choice = tk.StringVar(value='with_chemaxon')
        self.chemaxon_dir = tk.StringVar()
        self.sdf_path = tk.StringVar()
        self.property_file_path = tk.StringVar()
        self.sdf_id_field_name = tk.StringVar()
        self.property_field_name = tk.StringVar()
        self.compound_names = tk.StringVar(value='gen')

        chemaxon_frame = ttk.Labelframe(self, text='Descriptors (result in different interpretation)', name='chemaxon_frame')
        chemaxon_frame.grid(column=0, row=0, sticky=(tk.E, tk.W), columnspan=4, padx=5, pady=5)
        ttk.Radiobutton(chemaxon_frame, text='SiRMS (structural & physicochemical interpretation, Chemaxon required)', variable=self.descriptors_choice, value='with_chemaxon').grid(column=0, row=0, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='SiRMS (structural interpretation)', variable=self.descriptors_choice, value='no_chemaxon').grid(column=0, row=1, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='Atom-pair (structural interpretation)', variable=self.descriptors_choice, value='AP').grid(column=0, row=2, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='Atom-pair binary (structural interpretation)', variable=self.descriptors_choice, value='bAP').grid(column=1, row=2, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='Morgan2 (structural interpretation)', variable=self.descriptors_choice, value='MG2').grid(column=0, row=4, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='Morgan2 binary (structural interpretation)', variable=self.descriptors_choice, value='bMG2').grid(column=1, row=4, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='RDK (structural interpretation)', variable=self.descriptors_choice, value='RDK').grid(column=0, row=6, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='RDK binary (structural interpretation)', variable=self.descriptors_choice, value='bRDK').grid(column=1, row=6, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(chemaxon_frame, text='Topological torsion (structural interpretation)', variable=self.descriptors_choice, value='TT').grid(column=0, row=8, sticky=(tk.W), padx=5, pady=1)
        # ttk.Label(self, text='Optional. Path to cxcalc utility folder, '
        #                      'e.g. C:\\Program Files (x86)\\Chemaxon\\JChem\\bin').grid(column=0, row=0, sticky=(tk.W))
        # ttk.Entry(self, width=70, textvariable=self.chemaxon_dir).grid(column=0, row=1, sticky=(tk.W, tk.E))
        # ttk.Button(self, text='Browse...', command=self.__select_chemaxon_dir).grid(column=1, row=1, sticky=(tk.W))
        # ttk.Button(self, text='Auto-detect').grid(column=2, row=1, sticky=(tk.W))

        frame = ttk.Labelframe(self, text='SDF with compounds', name='sdf_label_frame')
        frame.grid(column=0, row=2, sticky=(tk.E, tk.W), columnspan=4, padx=5, pady=5)
        ttk.Label(frame, text='Path to SDF-file').grid(column=0, row=2, sticky=(tk.W, tk.S), padx=5)
        ttk.Label(frame, text='property field name').grid(column=2, row=2, sticky=(tk.W), padx=5)
        # ttk.Entry(frame, width=70, textvariable=self.sdf_path).grid(column=0, row=3, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))
        ttk.Combobox(frame, name='sdf_path_combobox', width=70, textvariable=self.sdf_path, values=self.__read_sdf_history()).grid(column=0, row=3, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))
        ttk.Button(frame, text='Browse...', command=self.__select_sdf_path).grid(column=1, row=3, sticky=(tk.W), padx=5, pady=(0, 5))
        ttk.Combobox(frame, name='property_field_name', width=20, textvariable=self.property_field_name, state='readonly').grid(column=2, row=3, sticky=(tk.W), padx=5, pady=(0, 5))

        frame = ttk.Labelframe(self, text='Optional. Compound names. External text file with compound property values', name='optional_label_frame')
        frame.grid(column=0, row=4, sticky=(tk.E, tk.W), columnspan=4, padx=5, pady=5)

        ttk.Radiobutton(frame, text='Automatically generate compound names', variable=self.compound_names, value='gen').grid(column=0, row=0, sticky=(tk.W), padx=5, pady=1)
        ttk.Radiobutton(frame, text='Use compound titles from SDF file', variable=self.compound_names, value='title').grid(column=0, row=1, sticky=(tk.W), padx=5, pady=1)

        inner_frame = ttk.Frame(frame, name='inner_frame')
        inner_frame.grid(column=0, row=3, columnspan=4, sticky=(tk.W, tk.E))
        ttk.Radiobutton(inner_frame, text='Use field values as compound names from SDF file', variable=self.compound_names, value='field').grid(column=0, row=3, sticky=(tk.W), padx=5, pady=1)
        cmbbox = ttk.Combobox(inner_frame, name='sdf_id_field_name', width=20, textvariable=self.sdf_id_field_name, state='readonly')
        cmbbox.grid(column=1, row=3, sticky=(tk.W), padx=5, pady=1)
        cmbbox.bind('<<ComboboxSelected>>', self.__set_compound_names_choice_event)
        # ttk.Entry(inner_frame, width=20, textvariable=self.sdf_id_field_name).grid(column=1, row=3, sticky=(tk.W), padx=5, pady=1)

        # ttk.Label(frame, text='Path to text file with property values').grid(column=0, row=5, sticky=(tk.W), padx=5, pady=(15, 0))
        # ttk.Entry(frame, width=70, textvariable=self.property_file_path).grid(column=0, row=7, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))
        # ttk.Button(frame, text='Browse...', command=self.__select_property_file_path).grid(column=1, row=7, sticky=(tk.W), padx=5, pady=(0, 5))

        self.models_frame = ModelsFrame(self, 'Models', 0, 10, True)

        self.sb_cpu_count = SpinboxCPUFrame(self, 0, 12)

        buttons_frame = ttk.Frame(self)
        buttons_frame.grid(column=0, row=20, columnspan=3, sticky=(tk.W, tk.E))

        ttk.Button(buttons_frame, text='Build models', command=self.__build_models).grid(column=0, row=0, sticky=(tk.E))
        ttk.Button(buttons_frame, text='Show statistics', command=self.__show_models_stat).grid(column=1, row=0, sticky=(tk.W))

        buttons_frame.columnconfigure(0, weight=1)
        buttons_frame.columnconfigure(1, weight=1)

        self.columnconfigure(0, weight=1)

        self.sdf_path.trace('w', self.__sdf_path_changed)
        self.property_field_name.trace('w', self.__prop_field_name_changed)
        self.descriptors_choice.trace('w', self.__descriptors_choice_changed)

        parent.add(self, text=tab_name)


class Tab_2(ttk.Frame):

    def get_frag_prefix(self):
        if self.frags_choice.get() != 'auto':
            return self.frags_choice.get()
        else:
            return self.auto_frags_choice.get().split(':')[0]

    def __select_user_frags(self):
        self.user_frags_path.set(filedialog.askopenfilename(filetypes=[('SMARTS file', '*.smarts; *.txt'),
                                                                       ('SMILES file', '*.smi; *.smiles'),
                                                                       ('SDF file', '*.sdf'),
                                                                       ('All files', '*.*')],
                                                            initialdir=os.path.dirname(self.master.children['tab_1'].sdf_path.get())))

    def __calc_contributions(self):

        """
        Intermediate files with atom ids and descriptors of fragmented molecules are stored in main dir for all
        pre-defined fragmentation schemes. Intermediate files generated during user-defined scheme are stored in
        the work (property) dir.
        Files with calculated contributions are stored in work (property) dir.
        """

        sdf_fname = self.master.children['tab_1'].sdf_path.get()[:-4] + '_std_lbl.sdf'

        prop_name = self.master.children['tab_1'].property_field_name.get()
        main_dir = os.path.dirname(sdf_fname)
        wd = os.path.realpath(os.path.join(main_dir, prop_name))

        # check for existence of XXX_frag_x.txt to avoid recalculation of fragments and descriptors
        if self.frags_choice.get() == 'user' and os.path.isfile(os.path.join(wd, 'user_frag_x.txt')):
            exist = messagebox.askyesnocancel(message='Do you want to keep existed user-defined fragmentation?', icon='question', default='yes')
            if exist is None:
                return
        else:
            exist = os.path.isfile(os.path.join(main_dir, self.get_frag_prefix() + '_frag_x.txt'))

        if not exist:

            if self.frags_choice.get() == 'atom':
                ids_fname = None
            elif self.frags_choice.get() == 'user':
                ids_fname = os.path.join(wd, self.get_frag_prefix() + '_frag_ids.txt')
            else:
                ids_fname = os.path.join(main_dir, self.get_frag_prefix() + '_frag_ids.txt')

            if self.frags_choice.get() == 'user':
                if self.user_frags_path.get() == '':
                    messagebox.showerror('Error!', 'Specify path to the fragments file.')
                    return
                else:
                    frag_fname = self.user_frags_path.get()
            elif self.frags_choice.get() == 'default':
                frag_fname = os.path.join(get_script_path(), 'default.smarts')

            # find atom ids
            if self.frags_choice.get() in ['user', 'default']:
                find_frags.main_params(in_sdf=sdf_fname,
                                       out_txt=ids_fname,
                                       in_frags=frag_fname,
                                       remove_all=False,
                                       verbose=True,
                                       error_fname=os.path.join(main_dir, "rdkit_fragmentation_errors.log"))
            # elif self.frags_choice.get() == 'rings':
            #     find_rings.main_params(in_sdf=sdf_fname,
            #                            out_txt=ids_fname,
            #                            verbose=True,
            #                            error_fname=os.path.join(main_dir, "indigo_errors.log"))
            elif self.frags_choice.get() == 'murcko':
                find_murcko.main_params(in_sdf=sdf_fname,
                                        out_txt=ids_fname,
                                        verbose=True,
                                        error_fname=os.path.join(main_dir, "rdkit_fragmentation_errors.log"))
            elif self.frags_choice.get() == 'auto':
                find_frags_auto.main_params(in_sdf=sdf_fname,
                                            out_txt=ids_fname,
                                            query=self.auto_schemes[self.auto_frags_choice.get()],
                                            max_cuts=3,
                                            keep_stereo=False,
                                            radius=[0],
                                            verbose=True,
                                            error_fname=os.path.join(main_dir, "rdkit_fragmentation_errors.log"))

            if self.frags_choice.get() == 'user':
                x_fname = os.path.join(wd, self.get_frag_prefix() + '_frag_x.txt')
            else:
                x_fname = os.path.join(main_dir, self.get_frag_prefix() + '_frag_x.txt')

            print("Descriptors calculation started. Please wait it can take some time")

            # SIRMS
            if self.master.children['tab_1'].descriptors_choice.get() in ['with_chemaxon', 'no_chemaxon']:

                # calc sirms descriptors
                if self.master.children['tab_1'].descriptors_choice.get() == 'with_chemaxon':
                    atom_diff = ['CHARGE', 'LOGP', 'HB', 'REFRACTIVITY']
                else:
                    atom_diff = ['elm']

                sirms.main_params(in_fname=sdf_fname,
                                  out_fname=x_fname,
                                  opt_diff=atom_diff,
                                  min_num_atoms=2,
                                  max_num_atoms=4,
                                  min_num_components=1,
                                  max_num_components=2,
                                  min_num_mix_components=2,
                                  max_num_mix_components=2,
                                  mix_fname=None,
                                  descriptors_transformation='num',
                                  mix_type='abs',
                                  opt_mix_ordered=False,
                                  opt_verbose=False,
                                  opt_noH=True,
                                  frag_fname=ids_fname,
                                  per_atom_fragments=True if ids_fname is None else False,
                                  self_association_mix=False,
                                  reaction_diff=False,
                                  quasimix=False,
                                  id_field_name=None,
                                  output_format='svm',
                                  ncores=min(10, max(self.master.children['tab_1'].sb_cpu_count.get_value(), 1)))


                # filter sirms descriptors
                filter_descriptors.main_params(in_fname=x_fname,
                                               out_fname=x_fname,
                                               file_format='svm')

            # RDKit
            else:
                descriptors.main_params(in_fname=sdf_fname,
                                        out_fname=x_fname,
                                        output_format='svm',
                                        get_fp=self.master.children['tab_1'].descriptors_choice.get(),
                                        opt_verbose=False,
                                        opt_noH=True,
                                        frag_fname=ids_fname,
                                        per_atom_fragments=True if ids_fname is None else False,
                                        id_field_name=None)

        else:

            # define path to descriptors of fragmented structures
            if self.frags_choice.get() == 'user':
                x_fname = os.path.join(wd, self.get_frag_prefix() + '_frag_x.txt')
            else:
                x_fname = os.path.join(main_dir, self.get_frag_prefix() + '_frag_x.txt')

        # calc contributions
        out_fname = os.path.join(wd, self.get_frag_prefix() + '_frag_contributions.txt')
        models = self.master.children['tab_1'].models_frame.get_selected_models()
        model_dir = os.path.join(wd, 'models')

        if self.master.children['tab_1'].descriptors_choice.get() == 'with_chemaxon':
            prop_name = ['overall', 'CHARGE', 'LOGP', 'HB', 'REFRACTIVITY']
        else:
            prop_name = ['overall']

        calc_frag_contrib.main_params(x_fname=x_fname,
                                      out_fname=out_fname,
                                      model_names=models,
                                      model_dir=model_dir,
                                      prop_names=prop_name,
                                      model_type=self.master.children['tab_1'].models_frame.model_type.get(),
                                      verbose=True,
                                      save_pred=True,
                                      input_format='svm',
                                      long_format=False,
                                      save_frag_ids=False,
                                      activity_file=os.path.join(os.path.dirname(self.master.children['tab_1'].sdf_path.get()),
                                                                 self.master.children['tab_1'].property_field_name.get().strip() + '.txt'))

        print("Calculation completed.")

    def __change_to_user(self, event):
        self.frags_choice.set(value='user')

    def __change_to_auto(self, event):
        self.frags_choice.set(value='auto')

    def __run_calc_contrib(self, event):
        self.__calc_contributions()

    def __init__(self, parent, tab_name):

        ttk.Frame.__init__(self, parent, name='tab_2')

        self.auto_schemes = OrderedDict()
        self.auto_schemes['auto1: [#6+0;!$(*=,#[!#6])]!@!=!#[*] (MMP like, quite exhaustive)'] = '[#6+0;!$(*=,#[!#6])]!@!=!#[*]'
        self.auto_schemes['auto2: [R]-[!R;!#1] (break bond between chain and ring atoms)'] = '[R]-[!R;!#1]'

        self.frags_choice = tk.StringVar(value='default')
        self.auto_frags_choice = tk.StringVar(value=list(self.auto_schemes.keys())[0])
        self.user_frags_path = tk.StringVar()

        frame = ttk.Labelframe(self, text='Select fragments set or fragmentation scheme')
        frame.grid(column=0, row=0, columnspan=3, sticky=(tk.E, tk.W), padx=5, pady=5)

        ttk.Radiobutton(frame, text='Default fragments', name='default_frag', value='default', variable=self.frags_choice).\
            grid(column=0, row=0, sticky=tk.W, padx=5, pady=(3, 1))
        # ttk.Radiobutton(frame, text='Ring systems', name='rings_frag', value='rings', variable=self.frags_choice).\
        #     grid(column=0, row=1, sticky=tk.W, padx=5, pady=(0, 1))
        # ttk.Radiobutton(frame, text='CCQ fragments', name='ccq_frag', value='ccq', variable=self.frags_choice, state='disabled').\
        #     grid(column=0, row=3, sticky=tk.W, padx=5, pady=(0, 1))
        # ttk.Radiobutton(frame, text='RECAP fragments', name='recap_frag', value='recap', variable=self.frags_choice, state='disabled').\
        #     grid(column=0, row=5, sticky=tk.W, padx=5, pady=(0, 1))
        ttk.Radiobutton(frame, text='Murcko scaffolds (frameworks)', name='murcko_frag',  value='murcko', variable=self.frags_choice).\
            grid(column=0, row=7, sticky=tk.W, padx=5, pady=(0, 1))
        ttk.Radiobutton(frame, text='Automatic fragmentation', name='auto_frag', value='auto', variable=self.frags_choice).\
            grid(column=0, row=8, sticky=tk.W, padx=5, pady=(0, 1))

        cmbbox = ttk.Combobox(frame, name='sdf_path_combobox', width=70, textvariable=self.auto_frags_choice, values=list(self.auto_schemes.keys()), state='readonly')
        cmbbox.grid(column=1, row=8, columnspan=2, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))

        ttk.Radiobutton(frame, text='User-defined fragments', name='user_frag', value='user', variable=self.frags_choice).\
            grid(column=0, row=9, sticky=tk.W, padx=5, pady=(0, 1))

        ttk.Radiobutton(frame, text='Per atom fragmentation (cannot be visualized by the program)', name='per_atom', value='atom', variable=self.frags_choice).\
            grid(column=0, row=20, sticky=tk.W, padx=5, pady=(0, 1))

        self.__entry_user_frags_path = ttk.Entry(frame, width=70, textvariable=self.user_frags_path)
        self.__entry_user_frags_path.grid(column=0, row=15, columnspan=2, sticky=(tk.W, tk.E), padx=5, pady=(0, 3))

        self.__btn_user_frags_path = ttk.Button(frame, text='Browse...', command=self.__select_user_frags)
        self.__btn_user_frags_path.grid(column=2, row=15, sticky=(tk.W), padx=5, pady=(0, 3))

        frame = tk.Frame(self)
        frame.grid(column=0, row=6, columnspan=2)

        self.__btn_calc_contrib = ttk.Button(frame, text='Calculate contributions', command=self.__calc_contributions)
        self.__btn_calc_contrib.grid(column=0, row=0)

        self.columnconfigure(0, weight=1)

        # events
        self.__entry_user_frags_path.bind('<Button-1>', self.__change_to_user)
        self.__entry_user_frags_path.bind('<Button-2>', self.__change_to_user)
        self.__entry_user_frags_path.bind('<Button-3>', self.__change_to_user)
        self.__entry_user_frags_path.bind('<Return>', self.__run_calc_contrib)
        self.__btn_user_frags_path.bind('<Button-1>', self.__change_to_user)
        cmbbox.bind('<<ComboboxSelected>>', self.__change_to_auto)
        cmbbox.bind('<Button-1>', self.__change_to_auto)
        cmbbox.bind('<Button-2>', self.__change_to_auto)
        cmbbox.bind('<Button-3>', self.__change_to_auto)

        parent.add(self, text=tab_name)


class Tab_3(ttk.Frame):

    def _show_models_list(self):

        model_names = self._get_model_list()

        if model_names is not None:

            model_frame = ttk.Labelframe(self, name='model_frame', text='Select models to visualise')
            model_frame.grid(column=1, row=0, sticky=(tk.EW, tk.N), padx=5, pady=5)

            self._models_chkbox = dict()
            for i, m in enumerate(model_names):
                self._models_chkbox[m] = tk.BooleanVar(value=True)
                ttk.Checkbutton(model_frame, name='chk_' + m, variable=self._models_chkbox[m], text=m.upper()).grid(column=0, row=i, sticky=(tk.W), padx=5, pady=(3, 1))

        else:
            if 'model_frame' in self.children.keys():
                self.children['model_frame'].destroy()

    def __run_plot(self, event):
        self.__plot_contributions()

    def __select_save_file_path(self):
        self.save_file_path.set(filedialog.asksaveasfilename(filetypes=[('PNG files', '*.png')],
                                                             defaultextension='.png',
                                                             initialdir=os.path.dirname(self.master.children['tab_1'].sdf_path.get())))

    def __get_selected_contributions(self):
        output = []
        if self.contr_charge.get(): output.append('CHARGE')
        if self.contr_logp.get(): output.append('LOGP')
        if self.contr_overall.get(): output.append('overall')
        if self.contr_hb.get(): output.append('HB')
        if self.contr_ref.get(): output.append('REFRACTIVITY')
        return output

    def _get_selected_models(self):
        output = [k for k, v, in self._models_chkbox.items() if v.get()]
        return output

    def _get_model_list(self):
        if self.master.children['tab_1'].sdf_path.get() != '':
            models_dir = os.path.join(os.path.dirname(self.master.children['tab_1'].sdf_path.get()),
                                      self.master.children['tab_1'].property_field_name.get(),
                                      "models")
            if os.path.isdir(models_dir):
                files = [f[:-4] for f in os.listdir(models_dir) if f.endswith('.pkl')]
                files.remove('scale')
                files.remove('bound_box')
                files.remove('var_names')
                return files
            else:
                return None
        else:
            return None

    def __plot_contributions(self):

        # at least one checkbox should be checked
        if not self.show_pic.get() and not self.save_pic.get():
            messagebox.showerror('ERROR!', 'You should check at least one option show figure on the screen or '
                                           'save it to the file')
            return

        # check and set output file name
        if self.save_pic.get():
            if self.save_file_path.get() == '':
                messagebox.showerror('ERROR!', 'Specify file name for the output png file')
                return
            else:
                fig_fname = self.save_file_path.get()
        else:
            fig_fname = None


        wd = os.path.join(os.path.dirname(self.master.children['tab_1'].sdf_path.get()), self.master.children['tab_1'].property_field_name.get())
        contr_fname = os.path.join(wd, self.master.children['tab_2'].get_frag_prefix() + '_frag_contributions.txt')

        # models = self.master.children['tab_1'].models_frame.get_selected_models()
        models = self._get_selected_models()

        plot_contributions.main_params(contr_fname=contr_fname,
                                       contr_names=self.__get_selected_contributions(),
                                       fig_fname=fig_fname,
                                       model_names=models,
                                       on_screen=self.show_pic.get(),
                                       min_M=int(self.filter_m.get()),
                                       min_N=int(self.filter_n.get()),
                                       model_type=self.master.children['tab_1'].models_frame.model_type.get())
        pass

    def __init__(self, parent, tab_name):

        ttk.Frame.__init__(self, parent, name='tab_3')

        self.contr_overall = tk.BooleanVar(value=True)
        self.contr_charge = tk.BooleanVar(value=False)
        self.contr_logp = tk.BooleanVar(value=False)
        self.contr_hb = tk.BooleanVar(value=False)
        self.contr_ref = tk.BooleanVar(value=False)

        frame = ttk.Labelframe(self, text='Select contribution types to visualise', name='contributions')
        frame.grid(column=0, row=0, sticky=(tk.EW, tk.N), padx=5, pady=5)

        # select properties
        overall_state = True
        if self.master.children['tab_1'].descriptors_choice.get() == 'with_chemaxon':
            charge_state, logp_state, hb_state, ref_state = 'normal', 'normal', 'normal', 'normal'
        else:
            charge_state, logp_state, hb_state, ref_state = 'disabled', 'disabled', 'disabled', 'disabled'

        ttk.Checkbutton(frame, variable=self.contr_overall, name='contr_overall', text='overall', state=overall_state).grid(column=0, row=0, sticky=(tk.W), padx=5, pady=(3, 1))
        ttk.Checkbutton(frame, variable=self.contr_charge, name='contr_charge', text='electrostatic (charge)', state=charge_state).grid(column=0, row=1, sticky=(tk.W), padx=5, pady=(0, 1))
        ttk.Checkbutton(frame, variable=self.contr_logp, name='contr_logp', text='hydrophobic (logp)', state=logp_state).grid(column=0, row=2, sticky=(tk.W), padx=5, pady=(0, 1))
        ttk.Checkbutton(frame, variable=self.contr_hb, name='contr_hb', text='hydrogen bonding (hb)', state=hb_state).grid(column=0, row=3, sticky=(tk.W), padx=5, pady=(0, 1))
        ttk.Checkbutton(frame, variable=self.contr_ref, name='contr_ref', text='dispersive (refractivity)', state=ref_state).grid(column=0, row=4, sticky=(tk.W), padx=5, pady=(0, 1))

        output_frame = ttk.Labelframe(self, text='Output options')
        output_frame.grid(column=0, row=5, columnspan=2, sticky=(tk.W, tk.E), padx=5, pady=5)

        self.filter_n = tk.StringVar(value='10')
        self.filter_m = tk.StringVar(value='10')

        ttk.Label(output_frame, text='Minimal occurrence of fragments (N) '.ljust(120, '.')).grid(column=0, row=0, sticky=tk.W, padx=5, pady=(0, 1))
        tk.Spinbox(output_frame, from_=1, to=1000000, textvariable=self.filter_n, width=10).grid(column=1, row=0, sticky=tk.W, padx=5, pady=(0, 1))
        ttk.Label(output_frame, text='Minimal number of molecules containing the same fragment (M) '.ljust(96, '.')).grid(column=0, row=1, sticky=tk.W, padx=5, pady=(0, 1))
        tk.Spinbox(output_frame, from_=1, to=1000000, textvariable=self.filter_m, width=10).grid(column=1, row=1, sticky=tk.W, padx=5, pady=(0, 1))

        self.show_pic = tk.BooleanVar(value=True)
        self.save_pic = tk.BooleanVar(value=True)
        self.save_file_path = tk.StringVar()

        ttk.Checkbutton(output_frame, variable=self.show_pic, text='Show figure in separate window').grid(column=0, row=2, sticky=(tk.W), padx=5, pady=(3, 1))
        ttk.Checkbutton(output_frame, variable=self.save_pic, text='Save figure to file').grid(column=0, row=3, sticky=(tk.W), padx=5, pady=(0, 1))

        self.__save_fig_file_path = ttk.Entry(output_frame, width=70, textvariable=self.save_file_path)
        self.__save_fig_file_path.grid(column=0, row=4, sticky=(tk.W, tk.E), padx=5, pady=(0, 1))
        ttk.Button(output_frame, text='Browse...', command=self.__select_save_file_path).grid(column=1, row=4, sticky=(tk.W), padx=5, pady=(0, 1))

        ttk.Button(self, text='Plot contributions', command=self.__plot_contributions).grid(column=0, row=10, columnspan=2, padx=5, pady=5)

        # events
        self.__save_fig_file_path.bind('<Return>', self.__run_plot)

        parent.add(self, text=tab_name)


class Tab_4(ttk.Frame):

    def __select_sdf_path(self):
        self.sdf_path_predict.set(filedialog.askopenfilename(filetypes=[('SDF files', '*.sdf')]))

    def __read_sdf_history(self):
        file_name = os.path.join(get_script_path(), 'history_pred.txt')
        if os.path.isfile(file_name):
            lines = open(file_name).readlines()
            lines = [line.strip() for line in lines]
            lines.reverse()
            return lines

    def __predict(self):

        if self.sdf_path_predict.get() == '':
            messagebox.showerror('ERROR!', 'Specify sdf filename and path.')
            return

        x_fname = os.path.splitext(self.sdf_path_predict.get())[0] + '_x.txt'

        # if x.txt does not exist then do all standardization and descriptors calculation
        use_x = False
        if os.path.isfile(x_fname):
            use_x = messagebox.askyesnocancel(message='It seems that folder with sdf file already contained calculated '
                                                      'descriptors - %s. Do you want to use this file as descriptors '
                                                      'for prediction?' % x_fname,
                                              icon='question', default='yes')
            if use_x is None:
                return

        project_dir = os.path.dirname(self.master.children['tab_1'].sdf_path.get())

        if not use_x:  # generate descriptors

            print("Descriptor calculation started")

            sdf_fname = self.sdf_path_predict.get()

            # SiRMS
            if self.master.children['tab_1'].descriptors_choice.get() in ['with_chemaxon', 'no_chemaxon']:

                # standardization and property labeling
                if self.master.children['tab_1'].descriptors_choice.get() == 'with_chemaxon':

                    # standardize
                    # print('Standardization is in progress...')
                    # std_sdf_tmp = self.sdf_path_predict.get() + '.std.sdf'
                    # run_params = ['standardize',
                    #               '-c',
                    #               quote_str(os.path.join(project_dir, 'std_rules.xml')),
                    #               quote_str(self.sdf_path_predict.get()),
                    #               '-f',
                    #               'sdf',
                    #               '-o',
                    #               quote_str(std_sdf_tmp)]
                    # call(' '.join(run_params), shell=True)

                    # calc atomic properties with Chemaxon
                    print('Atomic properties calculation is in progress...')
                    # input_sdf = self.sdf_path.get() if tmp_sdf is None else tmp_sdf
                    output_sdf = self.sdf_path_predict.get().rsplit(".")[0] + '_std_lbl.sdf'
                    calc_atomic_properties_chemaxon.main_params(in_fname=self.sdf_path_predict.get(),
                                                                out_fname=output_sdf,
                                                                prop=['charge', 'logp', 'acc', 'don', 'refractivity'],
                                                                pH=None,
                                                                cxcalc_path='cxcalc')

                    # os.remove(std_sdf_tmp)

                    atom_diff = ['CHARGE', 'LOGP', 'HB', 'REFRACTIVITY']
                    sdf_fname = output_sdf

                else:
                    atom_diff = ['elm']

                shutil.copyfile(os.path.join(get_script_path(), 'setup.txt'),
                                os.path.join(os.path.dirname(self.sdf_path_predict.get()), 'setup.txt'))

                sirms.main_params(in_fname=sdf_fname,
                                  out_fname=x_fname,
                                  opt_diff=atom_diff,
                                  min_num_atoms=2,
                                  max_num_atoms=4,
                                  min_num_components=1,
                                  max_num_components=2,
                                  min_num_mix_components=2,
                                  max_num_mix_components=2,
                                  mix_fname=None,
                                  descriptors_transformation='num',
                                  mix_type='abs',
                                  opt_mix_ordered=False,
                                  opt_verbose=False,
                                  opt_noH=True,
                                  frag_fname=None,
                                  per_atom_fragments=False,
                                  self_association_mix=False,
                                  reaction_diff=False,
                                  quasimix=False,
                                  id_field_name=None,
                                  output_format='svm',
                                  ncores=min(10, max(self.master.children['tab_1'].sb_cpu_count.get_value(), 1)))

            # RDKit
            else:
                descriptors.main_params(in_fname=sdf_fname,
                                        out_fname=x_fname,
                                        output_format='svm',
                                        get_fp=self.master.children['tab_1'].descriptors_choice.get(),
                                        opt_verbose=False,
                                        opt_noH=True,
                                        frag_fname=None,
                                        per_atom_fragments=False,
                                        id_field_name=None)

            print("Descriptor calculation finished")

        print("Prediction started")

        predict.main_params(x_fname=x_fname,
                            input_format='svm',
                            out_fname=os.path.splitext(self.sdf_path_predict.get())[0] + '_' + self.master.children['tab_1'].property_field_name.get().strip() + '_pred.txt',
                            model_names=self.master.children['tab_1'].models_frame.get_selected_models(),
                            model_dir=os.path.join(project_dir, self.master.children['tab_1'].property_field_name.get().strip(), 'models'),
                            model_type=self.master.children['tab_1'].models_frame.model_type.get(),
                            ad=['bound_box'],
                            verbose=False)

        print("Prediction finished")

    def __init__(self, parent, tab_name):

        ttk.Frame.__init__(self, parent, name='tab_4')

        self.sdf_path_predict = tk.StringVar()

        frame = ttk.Labelframe(self, text='SDF with compounds to predict', name='sdf_predict_label_frame')
        frame.grid(column=0, row=2, sticky=(tk.E, tk.W), columnspan=2, padx=5, pady=5)
        ttk.Label(frame, text='Path to SDF-file').grid(column=0, row=2, sticky=(tk.W, tk.S), padx=5)
        # ttk.Entry(frame, width=70, textvariable=self.sdf_path).grid(column=0, row=3, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))
        ttk.Combobox(frame, name='sdf_path_predict_combobox', width=70, textvariable=self.sdf_path_predict, values=self.__read_sdf_history()).grid(column=0, row=3, sticky=(tk.W, tk.E), padx=5, pady=(0, 5))
        ttk.Button(frame, text='Browse...', command=self.__select_sdf_path).grid(column=1, row=3, sticky=(tk.W), padx=5, pady=(0, 5))

        ttk.Label(self, text='*Predictions will be saved in the dir with the sdf-file specified above (<sdf_file_name>_<property_name>_pred.txt).').grid(column=0, row=7, sticky=(tk.W, tk.S), padx=5)
        ttk.Label(self, text='**Models for predictions will be taken from the first tab as well as descriptors').grid(column=0, row=8, sticky=(tk.W, tk.S), padx=5)

        ttk.Button(self, text='Predict', command=self.__predict).grid(column=0, row=10)

        parent.add(self, text=tab_name)


def main():

    root = tk.Tk()
    root.title("SPCI - structural and physico-chemical interpretation of QSAR models")

    content = ttk.Frame(root)

    lbl_copyright = ttk.Label(content, text='(c) Pavel Polishchuk 2014-2021 - v1.1.0')

    tabs = ttk.Notebook(content)
    tab_1 = Tab_1(tabs, 'Build models')
    tab_2 = Tab_2(tabs, 'Calc contributions')
    tab_3 = Tab_3(tabs, 'Plot contributions')
    tab_4 = Tab_4(tabs, 'Predict')

    content.grid(column=0, row=0, sticky=(tk.W, tk.N, tk.E, tk.S))
    tabs.grid(column=0, row=0, sticky=(tk.W, tk.N, tk.E, tk.S))
    lbl_copyright.grid(column=0, row=1, sticky=(tk.W, tk.E, tk.S))

    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    content.columnconfigure(0, weight=1)
    content.rowconfigure(0, weight=1)

    root.mainloop()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import os
import sys
import argparse
from datetime import datetime
from rdkit import Chem
from .find_frags_auto_rdkit import replace_no2


def load_smarts_file(fname):

    with open(fname) as f:
        frags = []
        for line in f:
            # skip comment lines and blank lines
            if line.strip().startswith('#') or not line.strip():
                continue
            tmp = line.strip().split("\t")
            if len(tmp) == 1 or (len(tmp) == 2 and tmp[1] == ""):
                frags.append({"name": tmp[0], "query": Chem.MolFromSmarts(tmp[0])})
            elif len(tmp) == 2:
                frags.append({"name": tmp[1], "query": Chem.MolFromSmarts(tmp[0])})
            elif len(tmp) == 3:
                frags.append({"name": tmp[1], "query": Chem.MolFromSmarts(tmp[0]),
                              "attached_atoms": list(map(int, tmp[2].split(",")))})
            else:
                print("Line " + line + " contains more than 3 tab-separated fields. Check file format.")
    return frags


def load_smi_fragments(fname):

    frags = []
    cansmi = set()

    with open(fname) as f:
        for line in f:
            tmp = line.strip().split()
            mol = Chem.MolFromSmiles(tmp[0])
            if mol is not None:
                smi = Chem.MolToSmiles(mol)
                if smi not in cansmi:
                    cansmi.add(smi)
                    frags.append({"name": smi, "query": mol})

    return frags


def load_sdf_fragments(fname):

    frags = []
    cansmi = set()

    for mol in Chem.SDMolSupplier(fname):
        if mol is not None:
            smi = Chem.MolToSmiles(mol)
            if smi not in cansmi:
                cansmi.add(smi)
                frags.append({"name": smi, "query": mol})

    return frags


def load_query_fragments(fname):

    frags = None

    if fname.endswith(".smi") or fname.endswith(".smiles"):
        frags = load_smi_fragments(fname)
    if fname.endswith(".sdf"):
        frags = load_sdf_fragments(fname)
    elif fname.endswith(".smart") or fname.endswith(".smarts") or fname.endswith(".txt"):
        frags = load_smarts_file(fname)
    else:
        print("Unsupported extension of fragments file. Use only smi or smiles for SMILES; "
              "sdf for SDF; smart, smarts, txt for SMARTS")

    return frags


def extend_atoms(mol, atom_ids, atom_numbers):

    def add_atoms(atom, atom_ids, atom_numbers):
        for nbr_atom in atom.GetNeighbors():
            if nbr_atom.GetIdx() not in atom_ids and nbr_atom.GetAtomicNum() in atom_numbers:
                atom_ids.append(nbr_atom.GetIdx())
                atom_ids = add_atoms(nbr_atom, atom_ids, atom_numbers)
        return atom_ids

    for atom in mol.GetAtoms():
        if atom.GetIdx() in atom_ids:
            atom_ids = add_atoms(atom, list(atom_ids), atom_numbers)

    return atom_ids


def get_map_atom_ids_list(mol, query, attached_atoms):
    """
    Returns list of tuples with atom ids if match and None otherwise
    ids started from 1, 2, 3, ...
    """

    m = Chem.AddHs(mol)
    ids = m.GetSubstructMatches(query)

    if attached_atoms is not None:
        ids = tuple(extend_atoms(m, i, attached_atoms) for i in ids)

    ids = tuple(tuple(sorted(j+1 for j in i)) for i in ids)   # 1-based ids 1, 2, 3, ...

    return ids if ids else None


def main_params(in_sdf, out_txt, in_frags, remove_all, verbose, error_fname):

    frags = load_query_fragments(in_frags)

    if frags is None:
        sys.exit()

    # find atoms ids which belongs to the fragments which should be removed
    # output example
    # {'paliperidone': {'hydroxy': [(12, 39)]},
    #  'salbutamol': {'hydroxy': [(16, 35), (13, 31), (15, 34)], 'amino': [(4, 24)]}, ...}

    with open(out_txt, "wt") as f:

        for i, mol in enumerate(Chem.SDMolSupplier(in_sdf, sanitize=False, removeHs=False)):
            if mol is not None:

                # modify representation of NO2 groups to charged version
                mol = replace_no2(mol)
                err = Chem.SanitizeMol(mol, catchErrors=True)
                if err != 0:
                    with open(error_fname, 'at') as f_err:
                        f_err.write('%s\t%s\t%s\n' % (datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                                      os.path.basename(__file__),
                                                      'Molecule %s failed to sanitize due to: ' % mol.GetProp("_Name") + str(err)))

                mol_name = mol.GetProp("_Name")
                if verbose:
                    print("Searching for fragments in", mol_name)

                for frag in frags:
                    ids = get_map_atom_ids_list(mol, frag["query"], frag.get("attached_atoms"))
                    if ids:
                        if remove_all:
                            tmp = set()
                            for v in ids:
                                tmp.update(v)
                            f.write(mol_name + "\t" + frag["name"] + "\t" + "\t".join(map(str, tuple(tmp))) + "\n")
                        else:
                            for v in ids:
                                f.write(mol_name + "\t" + frag["name"] + "\t" + "\t".join(map(str, v)) + "\n")

            else:
                print('Molecules #%i was skipped due to error' % i)
                with open(error_fname, 'at') as f_err:
                    f_err.write('%s\t%s\t%i\n' % (datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                                  os.path.basename(__file__), i + 1))


def main():

    parser = argparse.ArgumentParser(description='Find indices of fragments atoms in molecules from input sdf-file.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles.')
    parser.add_argument('-o', '--out', metavar='output.txt', required=True,
                        help='output text file; each line contains tab-separated fields: name of a molecule, '
                             'name of a fragment and list of corresponding atom numbers.')
    parser.add_argument('-f', '--frag', metavar='fragments.smarts', required=True,
                        help='file containing fragments in SMARTS(.txt|.smarts)/SMILES(.smi)/SDF(.sdf) format. '
                             'Tab-separated SMARTS/SMILES file contains 1) SMARTS/SMILES, 2) fragment name (optional), '
                             '3) comma-separated list of atom numbers for recursive addition (optional). '
                             'Comment lines start with #')
    parser.add_argument('-a', '--all', action='store_true', default=False,
                        help='if set this flag to true all identical fragments occurred more than once '
                             'in a molecule will be removed simultaneously.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='show progress on the screen.')
    parser.add_argument('-e', '--error_file', metavar='log_file_name.txt', default="indigo_errors.txt",
                        help='save names of molecules which cause error to a text log file. Default file name '
                             'indigo_errors.txt.')
    # parser.add_argument('-d', '--allow_duplicate_fragments', action='store_true', default=False,
    #                     help='pre-filtering of duplicate fragments. Default: false (duplicate fragments not allowed).')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_sdf = v
        if o == "out": out_txt = v
        if o == "frag": in_frags = v
        if o == "all": remove_all = v
        if o == "verbose": verbose = v
        if o == "error_file": error_fname = v
        # if o == "allow_duplicate_fragments": dupl = v

    main_params(in_sdf, out_txt, in_frags, remove_all, verbose, error_fname)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# author          : Pavel
# date            : 05.11.15
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel 2015
# license         : LGPLv3
#==============================================================================

import os
import numpy as np


mol_frag_sep = "###"


class SirmsFile():

    def __init__(self, fname, file_format='txt', frag_file=False, chunks=float("Inf")):

        if file_format == 'txt':
            self.__varnames = open(fname).readline().strip().split('\t')[1:]
            self.__mol_full_names = []           # keep the names of all molecules which were read
            self.__frag_full_names = []          # keep the names of all fragments which were read
                                                 # order is important since calculated contributions are saved in
                                                 # the same order (mol1###frag1#1 <- frag id is present)
            self.__is_mol_full_names_read = False
            self.__is_frag_full_names_read = False

        elif file_format == 'svm':
            self.__varnames = [v.strip() for v in open(os.path.splitext(fname)[0] + '.colnames').readlines()]
            self.__mol_full_names = [v.strip() for v in open(os.path.splitext(fname)[0] + '.rownames').readlines()]
            self.__frag_full_names = [v for v in self.__mol_full_names if v.find(mol_frag_sep) > -1]
            self.__is_frag_full_names_read = True

        self.__file = open(fname)
        self.__file_format = file_format
        self.__nlines = chunks               # file is read by chunks to avoid memory problems,
                                             # to read the whole file at once set to float('Inf')
        self.__is_frag_file = frag_file      # set which file to read ordinary (False) or with fragments (True)
                                             # in the latter case each chunk will contain all fragments for each
                                             # read molecules (another reading process)
        self.__cur_mol_read = 0              # number of mols (lines) already read (or index of line to read)

    def get_frag_full_names(self):
        return self.__frag_full_names

    def get_mol_names(self):
        return self.__mol_full_names

    def reset_read(self):
        self.__cur_mol_read = 0
        self.__file.seek(0)

    def __read_svm_next(self):
        if self.__cur_mol_read == len(self.__mol_full_names):
            return [], self.__varnames, np.asarray([])
        else:
            start = self.__cur_mol_read
            end = start + self.__nlines
            if end > len(self.__mol_full_names):
                end = len(self.__mol_full_names)
            else:
                cur_mol = self.__mol_full_names[end - 1].split(mol_frag_sep)[0]
                # end += 1
                while end < len(self.__mol_full_names) and self.__mol_full_names[end].split(mol_frag_sep)[0] == cur_mol:
                    end += 1
            # end -= 1
            lines = [self.__file.readline().strip() for _ in range(start, end)]
            x = []
            for line in lines:
                out = [0] * len(self.__varnames)
                for entry in line.split():
                    index, value = entry.split(':')
                    out[int(index)] = float(value)
                x.append(out)
            x = np.asarray(x)
            self.__cur_mol_read = end
            return self.__mol_full_names[start:end], self.__varnames, x

    def __read_txt_next(self):

        # skip header
        if self.__file.tell() == 0:
            self.__file.readline()

        lines = []

        # read first nlines
        i = 0
        while True:
            line = self.__file.readline()
            if not line:
                break
            lines.append(line.strip().split('\t'))
            i += 1
            if i >= self.__nlines:
                break

        # read next lines until meet a new molecule
        if self.__is_frag_file and lines:
            while True:
                cur_mol = lines[-1][0].split(mol_frag_sep)[0]
                cur_pos = self.__file.tell()
                line = self.__file.readline()
                if not line:
                    break
                if line.split('\t', 1)[0].split(mol_frag_sep)[0] == cur_mol:
                    lines.append(line.strip().split('\t'))
                else:
                    self.__file.seek(cur_pos)
                    break

        mol_names = []
        x = []
        for line in lines:
            mol_names.append(line[0])
            x.append(tuple(map(float, line[1:])))

        # add read frag names to the list (remain only mol name and frag name, e.g. mol1#frag1, not mol1#frag1#1)
        if not self.__is_frag_full_names_read:
            self.__frag_full_names.extend([mol_name for mol_name in mol_names if mol_name.find(mol_frag_sep) > -1])

        if not self.__is_mol_full_names_read:
            self.__mol_full_names.extend(mol_names)

        # if EOF then stop updating the list of fragments and mol names (it is read only once)
        if len(mol_names) < self.__nlines:
            self.__is_frag_full_names_read = True
            self.__is_mol_full_names_read = True

        return mol_names, self.__varnames, np.asarray(x)

    def read_next(self):
        if self.__file_format == 'txt':
            return self.__read_txt_next()
        elif self.__file_format == 'svm':
            return self.__read_svm_next()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import argparse
import matplotlib.pyplot as plt
from collections import defaultdict, Counter
import numpy as np

mol_frag_sep = "###"


def read_contrib_file(fname, model_names, min_M, min_N, contr_names):
    """
    OUTPUT
    {'gbm': {'charge': [0.119, 0.174, 0.083, 0.220, ... ],
            {'overall': [0.201, 0.256, 0.283, 0.420, ... ],
              ...
            },
     'rf': ...}
    """
    d = defaultdict(dict)

    with open(fname) as f:

        names = f.readline().strip().split('\t')[1:]

        # count number of molecules for each fragment
        frag_mol_count = defaultdict(int)
        frag_names = []

        for n in names:
            mol_name, frag_name = n.split(mol_frag_sep)
            frag_mol_count[frag_name] += 1
            frag_names.append(frag_name)

        # count number of each fragment
        frag_count = Counter(frag_names)

        # create new fragment names and create list of filtered indices (according to min_M and min_N)
        keep_ids = []
        for i, v in enumerate(frag_names):
            frag_names[i] = frag_names[i] + " (M=" + str(frag_mol_count[v]) + ", N=" + str(frag_count[v]) + ")"
            if frag_mol_count[v] >= min_M and frag_count[v] >= min_N:
                keep_ids.append(i)

        # filter out frag_names by keep_ids
        frag_names = [frag_names[i] for i in keep_ids]

        for line in f:
            tmp = line.strip().split('\t')
            model_name, prop_name = tmp[0].rsplit("_", 1)
            # skip contributions which are not selected
            if prop_name not in contr_names:
                continue
            if "all" in model_names or model_name in model_names:
                values = list(map(float, tmp[1:]))
                # filter out values by keep_ids
                values = [values[i] for i in keep_ids]
                d[model_name][prop_name] = values

    return frag_names, d


def median(mylist):
    sorts = sorted(mylist)
    length = len(sorts)
    if not length % 2:
        return (sorts[length // 2] + sorts[length // 2 - 1]) / 2.0
    return sorts[length // 2]


def add_consensus_average(contr_dict):
    consensus_dict = {}
    for model_name in contr_dict.keys():
        for prop_name in contr_dict[model_name].keys():
            if prop_name not in consensus_dict.keys():
                consensus_dict[prop_name] = [0] * len(contr_dict[model_name][prop_name])
            consensus_dict[prop_name] = [a + b for (a,b) in zip(consensus_dict[prop_name],
                                                                contr_dict[model_name][prop_name])]
    for prop_name in consensus_dict:
        consensus_dict[prop_name] = [v / len(contr_dict) for v in consensus_dict[prop_name]]
    contr_dict["consensus"] = consensus_dict


def prep_data_boxplot(contr_dict, frag_names_list, add_consensus=True):
    """
    INPUT
    {'gbm': {'charge': [0.119, 0.174, 0.083, 0.220, ... ],
             'overall': [0.201, 0.256, 0.283, 0.420, ... ],
              ...
            },
     'rf': ...}

    OUTPUT
    fragment_names_list = ['SO2', 'NO2', ...]  - corresponds to number of rows
    {'gbm': {'charge': [[0.119, 0.174, 0.083, 0.220, ... ],
                        [0.119, 0.174, 0.083, 0.220, ... ],
                        ...
                       ],
            {'overall': [[0.201, 0.256, 0.283, 0.420, ... ],
                         [0.429, 0.345, 0.121, 0.301, ... ],
                         ...
                        ],
     'rf': ...}
    """
    if add_consensus and len(contr_dict) > 1:
        add_consensus_average(contr_dict)
    output_dict = dict()
    for model_name in contr_dict.keys():
        output_dict[model_name] = dict()
        for prop_name in contr_dict[model_name]:
            tmp = defaultdict(list)
            for i, value in enumerate(contr_dict[model_name][prop_name]):
                tmp[frag_names_list[i]].append(value)
            sorted_frag_names = sorted(tmp.keys())
            output_dict[model_name][prop_name] = [tmp[n] for n in sorted_frag_names]
    return sorted_frag_names, output_dict


def sorted_ids(ls, ascending=True):
    if ascending:
        ids = sorted(range(len(ls)), key=lambda k: ls[k])
    else:
        ids = sorted(range(len(ls)), key=lambda k: -ls[k])
    ids = sorted(enumerate(ids), key=lambda x: x[1])
    ids = [i for i, j in ids]
    return ids


def sort_by_median(list_of_lists, ascending=True):
    m = [median(i) for i in list_of_lists]
    ids = sorted_ids(m, ascending)
    return ids


def prep_data_barplot(contr_dict, frag_names_list, add_consensus=True):
    """
    INPUT
    {'gbm': {'charge': [0.119, 0.174, 0.083, 0.220, ... ],
             'overall': [0.201, 0.256, 0.283, 0.420, ... ],
              ...
            },
     'rf': ...}

    OUTPUT
    fragment_names_list = ['SO2', 'NO2', ...]  - corresponds to number of rows
    {'gbm': {'charge': [0.119, 0.219, ... ],   - median contribution values
            {'overall': [0.119, 0.219, ... ],
                         ...
                        ],
     'rf': ...}
    """
    sorted_frag_names, d = prep_data_boxplot(contr_dict, frag_names_list, add_consensus)
    for model_name in d.keys():
        for prop_name in d[model_name].keys():
            d[model_name][prop_name] = [median(v) for v in d[model_name][prop_name]]
    return sorted_frag_names, d


def prep_data_barplot_binary_class(contr_dict, frag_names_list, add_consensus=True):
    """
    INPUT
    {'gbm': {'charge': [0.119, 0.174, 0.083, 0.220, ... ],
             'overall': [0.201, 0.256, 0.283, 0.420, ... ],
              ...
            },
     'rf': ...}

    OUTPUT
    fragment_names_list = ['SO2', 'NO2', ...]  - corresponds to number of rows
    {'gbm': {'charge': [(-2, 8), (-12, 3), ... ],   - negative and positive contribution values (sum of all negatives and sum of all positives)
            {'overall': [(0, 10), (-11, 1), ... ],
                         ...
                        ],
     'rf': ...}
    """
    sorted_frag_names, d = prep_data_boxplot(contr_dict, frag_names_list, add_consensus)
    for model_name in d.keys():
        for prop_name in d[model_name].keys():
            tmp = []
            for v in d[model_name][prop_name]:
                n = sum([vv for vv in v if vv < 0])
                p = sum([vv for vv in v if vv > 0])
                tmp.append((n, p))
            d[model_name][prop_name] = tmp
    return sorted_frag_names, d


def main_params(contr_fname, contr_names, fig_fname, model_names, on_screen, min_M, min_N, model_type):

    frag_names, contr_dict = read_contrib_file(contr_fname, model_names, min_M, min_N, contr_names)

    if model_type == 'reg' or model_type == 'class':

        # plot boxplot for single property (contribution)
        if len(contr_names) == 1:

            sorted_frag_names, data = prep_data_boxplot(contr_dict, frag_names)

            if "consensus" in data.keys():
                x_label_position = sort_by_median(data["consensus"][contr_names[0]])
            else:
                x_label_position = sort_by_median(list(data.values())[0][contr_names[0]])

            fig = plt.figure(1)
            fig.suptitle(contr_names[0] + " contributions", fontsize='x-large', color='blue')

            for i, model_name in enumerate(sorted(data.keys(), reverse=True)):

                ax = fig.add_subplot(1, len(data), len(data) - i)
                ax.grid(True, axis='y')
                plt.title(model_name)
                bp = ax.boxplot(x=list(data[model_name][contr_names[0]]),
                                sym="o",
                                positions=x_label_position,
                                vert=False)
                ax.axvline(0, color='grey', linestyle='--')

                # show x labels only on the first plot
                if i == len(data) - 1:
                    ax.set_yticklabels(sorted_frag_names)
                else:
                    ax.set_yticklabels([])
                for box in bp["boxes"]:
                    box.set(color="blue")
                for w in bp["whiskers"]:
                    w.set(color="blue", linestyle="-")
                for flier in bp["fliers"]:
                    flier.set(markeredgecolor="grey",
                              markeredgewidth=0,
                              markerfacecolor="grey",
                              markersize=4)

            # save figure if specified
            # fig.subplots_adjust(bottom=0, left=0, top=1, right=1)
            # fig.tight_layout()
            if fig_fname is not None:
                w = 3.5 * len(data) + 3.5
                h = 0.5 * len(sorted_frag_names) + 1
                fig.set_size_inches(w, h)
                fig.subplots_adjust(wspace=0.05, bottom=0.01, left=0.01, top=(h-1)/h, right=0.99)
                fig.savefig(fig_fname, dpi=300, bbox_inches='tight')
                fig.set_size_inches(8, 6)
            if on_screen:
                plt.show()
            plt.clf()

        # plot bar plot with medians for several properties (contributions)
        else:
            sorted_frag_names, data = prep_data_barplot(contr_dict, frag_names)

            # remove overall to avoid its plotting
            # for v in data.values():
            #     del v["overall"]

            fig = plt.figure(1)
            bar_width = 0.8 / len(list(data.values())[0])
            colors = ['royalblue', 'tomato', 'yellowgreen', 'darkturquoise', 'darkorchid']

            for i, model_name in enumerate(sorted(data.keys(), reverse=True)):
                ax = fig.add_subplot(1, len(data), len(data) - i)
                plt.title(model_name)
                ind = np.arange(len(sorted_frag_names))
                ax.set_yticks(ind + 0.4)
                ax.grid(True, axis='y')

                for j, prop_name in enumerate(data[model_name].keys()):
                    br = ax.barh(bottom=ind + j * bar_width,
                                 width=data[model_name][prop_name],
                                 height=bar_width,
                                 color=colors[j])
                    if prop_name == "CHARGE":
                        br.set_label("electrostatic")
                    if prop_name == "LOGP":
                        br.set_label("hydrophobic")
                    if prop_name == "HB":
                        br.set_label("hydrogen bonding")
                    if prop_name == "REFRACTIVITY":
                        br.set_label("dispersive")
                    if prop_name == "overall":
                        br.set_label("overall")

                if i == len(data) - 1:
                    # reverse the order
                    handles, labels = ax.get_legend_handles_labels()
                    ax.legend(handles[::-1], labels[::-1], bbox_to_anchor=(0, 1.12),
                              ncol=1, mode="expand", borderaxespad=0., fontsize='medium')
                    ax.set_yticklabels(sorted_frag_names, rotation=0)
                else:
                    ax.set_yticklabels([])

            fig.tight_layout()
            if fig_fname is not None:
                w = 3.5 * len(data) + 3.5
                h = 0.5 * len(sorted_frag_names) + 1
                fig.set_size_inches(w, h)
                fig.subplots_adjust(wspace=0.05, bottom=0.01, left=0.01, top=(h-1)/h, right=0.99)
                fig.savefig(fig_fname, dpi=300, bbox_inches='tight')
                fig.set_size_inches(8, 6)
            if on_screen:
                plt.show()
            plt.clf()
                
    elif model_type == 'class':  # this is experimental code for integer contributions -1, 0, 1
                                 # it will be never executed

        sorted_frag_names, data = prep_data_barplot_binary_class(contr_dict, frag_names)

        fig = plt.figure(1)
        bar_width = 0.8 / len(list(data.values())[0])
        colors = ['royalblue', 'tomato', 'yellowgreen', 'darkturquoise', 'darkorchid']

        ind = np.arange(len(sorted_frag_names))

        for i, model_name in enumerate(sorted(data.keys(), reverse=True)):
            ax = fig.add_subplot(1, len(data), len(data) - i)
            plt.title(model_name)
            ax.set_yticks(ind + 0.4)
            ax.grid(True, axis='y')

            for j, prop_name in enumerate(data[model_name].keys()):
                br = ax.barh(bottom=ind + j * bar_width,
                             width=[i[1] for i in data[model_name][prop_name]],
                             height=bar_width,
                             color=colors[j])
                if prop_name == "CHARGE":
                    br.set_label("electrostatic")
                if prop_name == "LOGP":
                    br.set_label("hydrophobic")
                if prop_name == "HB":
                    br.set_label("hydrogen bonding")
                if prop_name == "REFRACTIVITY":
                    br.set_label("dispersive")
                if prop_name == "overall":
                    br.set_label("overall")
                br = ax.barh(bottom=ind + j * bar_width,
                             width=[i[0] for i in data[model_name][prop_name]],
                             height=bar_width,
                             color=colors[j])

            if i == len(data) - 1:
                # reverse the order
                handles, labels = ax.get_legend_handles_labels()
                ax.legend(handles[::-1], labels[::-1], bbox_to_anchor=(0, 1.12),
                          ncol=1, mode="expand", borderaxespad=0., fontsize='medium')
                ax.set_yticklabels(sorted_frag_names, rotation=0)
            else:
                ax.set_yticklabels([])

        fig.tight_layout()
        if fig_fname is not None:
            fig.set_size_inches(3.5 * len(data) + 3.5, 0.5 * len(sorted_frag_names) + 1)
            fig.subplots_adjust(wspace=0.05)
            fig.savefig(fig_fname, dpi=300, bbox_inches='tight')
            fig.set_size_inches(8, 6)
        if on_screen:
            plt.show()


def main():

    parser = argparse.ArgumentParser(description='Plot fragments contributions.')
    parser.add_argument('-i', '--input', metavar='frag_contributions.txt', required=True,
                        help='input text file with calculated fragment contributions.')
    parser.add_argument('-o', '--output', metavar='output.png', default=None,
                        help='output file name for plot.')
    parser.add_argument('-c', '--contributions', metavar='[overall charge logp hb refractivity]', default=['overall'],
                        nargs='*', help='name of contributions to calculate.')
    parser.add_argument('-m', '--models', metavar='[rf gbm svr pls knn]', default=["all"], nargs='*',
                        help='file name of saved models (without extension)')
    parser.add_argument('-s', '--show_in_separate_window', action='store_true', default=False,
                        help='show plot on the screen in a  separate window')
    parser.add_argument('-N', '--min_fragments_count', metavar='', default=1,
                        help='minimal number of fragments in the dataset compounds')
    parser.add_argument('-M', '--min_molecules_count', metavar='', default=1,
                        help='minimal number of compounds containing a fragment')
    parser.add_argument('-t', '--model_type', metavar='reg|class', default='reg',
                        help='models type: reg for regression and class for classification')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "input": contr_fname = v
        if o == "contributions": contr_names = v
        if o == "output": fig_fname = v
        if o == "models": model_names = v
        if o == "show_in_separate_window": on_screen = v
        if o == "min_fragments_count":
            try:
                min_N = int(v)
            except ValueError:
                print("Incorrect minimal number of fragments. Not a number. Default value (1) will be used.")
                min_N = 1
        if o == "min_molecules_count":
            try:
                min_M = int(v)
            except ValueError:
                print("Incorrect minimal number of molecules containing a fragment. Not a number. "
                      "Default value (1) will be used.")
                min_M = 1
        if o == "model_type": model_type = v

    main_params(contr_fname, contr_names, fig_fname, model_names, on_screen, min_M, min_N, model_type)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
from collections import OrderedDict
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem
from rdkit.Chem.AtomPairs import Pairs, Torsions

from sirms.files import SvmSaver, LoadFragments
from sirms.sirms import SaveSimplexes


mol_frag_sep = "###"


def CalcMolFP(m, i, opt_noH, f, frags=None, per_atom_fragments=None, id_field_name=None):

    def get_fp_as_dict(mol, opt_noH, f):
        """calc specified fingerprint for input molecule and return nonzero elements of it as dict
        """
        
        if opt_noH:
            # Chem.RemoveHs(mol)  # !it doesnt help, anyway next line gets them Hs back
            mol = Chem.RWMol(mol)
            for idx in reversed(range(mol.GetNumAtoms())):  # reverse because ids of atoms change while iter
                if mol.GetAtomWithIdx(idx).GetAtomicNum() == 1:
                    mol.RemoveAtom(idx)
            Chem.FastFindRings(mol)  # needs to calc morganfp, otherwise err "no ringinfo"
        if f in (GetMorganFingerprint_2, Pairs.GetAtomPairFingerprint, Torsions.GetTopologicalTorsionFingerprint,
                 Get_RDKFP_24):
            
            return {str(k): v for k, v in f(mol).GetNonzeroElements().items()}  # keys must be str input to saver, val-int
        elif f in [GetMorganFingerprint_2_bin, Pairs.GetAtomPairFingerprintAsBitVect, Get_RDKFP_24_bin]:
            return {str(k): int(v) for k, v in enumerate(DataStructs.BitVectToText(f(mol))) if v == '1'}

    mol_dict = OrderedDict()
    if id_field_name is not None:
        nm = m.GetProp(id_field_name)
    elif m.GetProp("_Name") == "":
        nm = 'auto_generated_id_' + str(i + 1)  # 1-based as in sirms.py
    else:
        nm = m.GetProp("_Name")
    res = get_fp_as_dict(m, opt_noH, f)
    if res:  # if all descriptors for a molecule are zero skip fragmentation (NC(=O)N gives no TT, removal of C(=O)N from it gives TT)
        mol_dict[nm] = res
        if per_atom_fragments:
            counter = 0
            for idx in range(m.GetNumAtoms()):
                if m.GetAtomWithIdx(idx).GetAtomicNum() > 1:
                    rw_m = Chem.RWMol(m)
                    rw_m.GetAtoms()[idx].SetAtomicNum(0)
                    mol_dict[nm + mol_frag_sep + str(idx + 1) + "#" + str(counter)] = get_fp_as_dict(rw_m, opt_noH, f)
                    counter += 1
        elif frags and nm in frags:
            for k, v in frags[nm].items():
                rw_m = Chem.RWMol(m)
                for idx in sorted(v, reverse=True):  # note we don't check if atom== H (is it ok?)
                    rw_m.GetAtoms()[idx-1].SetAtomicNum(0)
                mol_dict[nm + mol_frag_sep + k] = get_fp_as_dict(rw_m, opt_noH, f)
    return mol_dict


def main_params(in_fname, out_fname, output_format, get_fp, opt_verbose, opt_noH, frag_fname,
                per_atom_fragments, id_field_name):

    funcs = {'bMG2': GetMorganFingerprint_2_bin,
             'MG2': GetMorganFingerprint_2,
             'bAP': Pairs.GetAtomPairFingerprintAsBitVect,
             'AP': Pairs.GetAtomPairFingerprint,
             'TT': Torsions.GetTopologicalTorsionFingerprint,
             'bRDK': Get_RDKFP_24_bin,
             'RDK': Get_RDKFP_24}
    get_fp = funcs[get_fp]

    # load sdf and get dict of fp (like sirms dict)
    input_file_extension = in_fname.strip().split(".")[-1].lower()
    if input_file_extension == 'sdf':
        saver = None
        mols = None
        if output_format == "svm":
            saver = SvmSaver(out_fname)
        if output_format == "txt":
            mols = OrderedDict()  # key - molname, val- mol; if frags: key - molname or mol+fragname, val-mol for mol or part b
        frags = LoadFragments(frag_fname)
        for i, m in enumerate(Chem.SDMolSupplier(in_fname, removeHs=False)):
            if m is not None:
                res = CalcMolFP(m, i, opt_noH=opt_noH, f=get_fp, frags=frags, per_atom_fragments=per_atom_fragments,
                                id_field_name = id_field_name)
                if output_format == "txt":
                    mols.update(res)
                if output_format == "svm":
                    for mol_name, descr_dict in res.items():
                        saver.save_mol_descriptors(mol_name, descr_dict)
        if output_format == "txt":
            SaveSimplexes(out_fname, mols, output_format)
    else:
        print("Input file extension should be SDF Current file has %s. Please check it." %
              input_file_extension.upper())
        return None


def GetMorganFingerprint_2(m):
    return AllChem.GetMorganFingerprint(m, radius=2)


def GetMorganFingerprint_2_bin(m):
    return AllChem.GetMorganFingerprintAsBitVect(m, radius=2)


def Get_RDKFP_24(m):
    return AllChem.UnfoldedRDKFingerprintCountBased(m, minPath=2, maxPath=4)


def Get_RDKFP_24_bin(m):
    return AllChem.RDKFingerprint(m, minPath=2, maxPath=4)


def entry_point():
    parser = argparse.ArgumentParser(description='Calculate fingerprint descriptors')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=False, default=None,
                        help='input file ( sdf with standardized structures')
    parser.add_argument('-o', '--out', metavar='output.txt', required=False, default=None,
                        help='output file with calculated descriptors. Can be in text or sparse svm format.')
    parser.add_argument('-b', '--output_format', metavar='output_format', default='txt',
                        help='format of output file with calculated descriptors (txt|svm). '
                             'Txt is ordinary tab-separated text file. Svm is sparse format, two additional files will '
                             'be saved with extensions .colnames and .rownames. Default: txt.')
    parser.add_argument('--fp_type', required=True,
                        help='Type of fingerprints to calculate choose one of: MG2 or bMG2 (Morgan radius 2), '
                             'AP or bAP (atom-pair), RDK or bRDK (2-4 atoms RDK fingerprint), '
                             'TT (topological torsion). Prefix b means binary fingerprint of length 2048.')
    parser.add_argument('-x', '--noH', action='store_true', default=False,
                        help='if set this flag hydrogen atoms will be excluded from the simplexes calculation.')
    parser.add_argument('-f', '--fragments', metavar='fragments.txt', default=None,
                        help='text file containing list of names of single compounds, fragment names and atom '
                             'indexes of fragment to remove (all values are tab-separated).')
    parser.add_argument('--per_atom_fragments', action='store_true', default=False,
                        help='if set this flag input fragments will be omitted and single atoms will be considered '
                             'as fragments.')
    parser.add_argument('-w', '--id_field_name', metavar='field_name', default=None,
                        help='field name of unique ID for compounds. '
                             'If omitted for sdf molecule titles will be used or auto-generated names')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='if set this flag progress will be printed out (may cause decrease in speed).')

    args = vars(parser.parse_args())

    for o, v in args.items():
        if o == "in": in_fname = v
        if o == "out": out_fname = v
        if o == "output_format": output_format = v
        if o == "fp_type": get_fp = v
        if o == "verbose": opt_verbose = v
        if o == "noH": opt_noH = v
        if o == "fragments": frag_fname = v
        if o == "per_atom_fragments": per_atom_fragments = v
        if o == "id_field_name": id_field_name = v

    main_params(in_fname=in_fname, out_fname=out_fname, output_format=output_format, get_fp=get_fp,
                opt_verbose=opt_verbose, opt_noH=opt_noH, frag_fname=frag_fname,
                per_atom_fragments=per_atom_fragments, id_field_name=id_field_name)


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import os
import argparse
from subprocess import call


def quote_str(s):
    return '"%s"' % s


def add_pH(params, pH):
    if pH is not None:
        params.extend(["-H", pH])
    return (params)


def add_HB(molstr):

    for i, line in enumerate(molstr):
        if line.strip() == ">  <ACC>":
            acc = ["I" if el == "0" else "A" for el in molstr[i+1].strip().split(";")]
        if line.strip() == ">  <DON>":
            don = ["I" if el == "0" else "D" for el in molstr[i+1].strip().split(";")]

    hb = []
    for a, d in zip(acc, don):
        if a == "A" and d == "D":
            hb.append("A|D")
        elif a == "A":
            hb.append("A")
        elif d == "D":
            hb.append("D")
        else:
            hb.append("I")

    molstr.insert(len(molstr) - 1, ">  <HB>\n")
    molstr.insert(len(molstr) - 1, ";".join(hb) + "\n")
    molstr.insert(len(molstr) - 1, "\n")

    return molstr


def main_params(in_fname, out_fname, prop, pH, cxcalc_path):

    start_params = [cxcalc_path]   # remove quotes because Win cannot run "cxcalc" only "dir/cxcalc"
    start_params.extend(["-o", quote_str(out_fname)])
    start_params.append("-S")
    log_fname = os.path.splitext(out_fname)[0] + "_cxcalc.log"
    start_params.extend(["--log", quote_str(log_fname)])
    start_params.append(quote_str(in_fname))

    # logp, charge, atom_polarizability, refractivity, acc, don
    run_params = start_params[:]
    if "logp" in prop:
        run_params.extend(["logp", "-t", "increments"])
        run_params = add_pH(run_params, pH)
    if "charge" in prop:
        run_params.append("charge")
        run_params = add_pH(run_params, pH)
    if "atompol" in prop:
        run_params.append("atompol")
        run_params = add_pH(run_params, pH)
    if "refractivity" in prop:
        run_params.extend(["refractivity", "-t", "increments"])
    if "acc" in prop:
        run_params.append("acc")
        run_params = add_pH(run_params, pH)
    if "don" in prop:
        run_params.append("don")
        run_params = add_pH(run_params, pH)
    if run_params != start_params:
        call(' '.join(run_params), shell=True)

    # read numbers of molecules which produce errors
    mol_errors = []
    if os.path.isfile(log_fname):
        with open(log_fname) as f:
            for line in f:
                if line.find("<_MOLCOUNT>") >= 0:
                    mol_errors.append(int(f.readline().strip()))

    # create new field HB - hydrogen bond parameters
    # A - acceptor
    # D - donor
    # AD - acceptor & donor
    # I - neither acceptor nor donor
    # and remove error molecules

    tmp_out_fname = out_fname + ".tmp"
    add_hb = "acc" in prop and "don" in prop
    if add_hb or mol_errors:
        with open(tmp_out_fname, "wt") as fout:
            with open(out_fname) as fin:

                molstr = []
                cur_mol_id = 1
                for line in fin:
                    if line.rstrip() != "$$$$":
                        molstr.append(line)
                    else:
                        molstr.append(line)
                        if mol_errors and cur_mol_id in mol_errors:
                            molstr = []
                        elif add_hb:
                            molstr = add_HB(molstr)
                            fout.writelines(molstr)
                            molstr = []
                        cur_mol_id += 1

        os.remove(out_fname)
        os.rename(tmp_out_fname, out_fname)


def main():

    parser = argparse.ArgumentParser(description='Calculate atomic properties with Chemaxon cxcalc tool. '
                                                 'Molecules which produce errors are omitted from output file. '
                                                 'Omitted molecules will be listed in the created log file.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles')
    parser.add_argument('-o', '--out', metavar='output.sdf', required=True,
                        help='output sdf file with added calculated atomic properties')
    parser.add_argument('-p', '--properties', metavar='', nargs='*',
                        default=['charge', 'logp', 'acc', 'don', 'refractivity'],
                        help="list of atomic properties to calc. Default: 'charge', 'logp', 'acc', 'don', "
                             "'refractivity'. If 'acc' and 'don' specified simultaneously hydrogen bonding "
                             "descriptors (hb) will be generated.")
    parser.add_argument('-H', '--pH', metavar='pH_value', default=None,
                        help='pH value which will be used during calculation of atomic properties')
    parser.add_argument('-c', '--cxcalc_path', metavar='path-to-cxcalc.bat',
                        default="cxcalc",
                        help='path to cxcalc executable file. It is usually '
                             'c:\\Program Files (x86)\\ChemAxon\\JChem\\bin\\cxcalc on Windows')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_fname = v
        if o == "out": out_fname = v
        if o == "properties": prop = v
        if o == "pH": pH = v
        if o == "cxcalc_path": cxcalc_path = v

    main_params(in_fname, out_fname, prop, pH, cxcalc_path)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import os
import argparse
import numpy as np
import sklearn.model_selection as ms
import warnings

from time import strftime
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor, RandomForestClassifier, \
    GradientBoostingClassifier
from sklearn.cross_decomposition import PLSRegression
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
import joblib
from sklearn.preprocessing import StandardScaler
from sklearn import svm
from multiprocessing import cpu_count
from scipy.sparse import dok_matrix

from .predict import get_major_vote


def load_sirms_txt(fname, names=None):
    if names:
        names = set(names)
    with open(fname) as f:
        descr_names = f.readline().strip().split("\t")[1:]
        case_names = []
        x = []
        for line in f:
            tmp = line.strip().split("\t")
            if not names or tmp[0] in names:
                case_names.append(tmp[0])
                x.append(tuple(map(float, tmp[1:])))

    return descr_names, case_names, np.asarray(x)


def load_sirms_svm(fname, names=None):
    descr_names = [v.strip() for v in open(os.path.splitext(fname)[0] + '.colnames').readlines()]
    case_names = [v.strip() for v in open(os.path.splitext(fname)[0] + '.rownames').readlines()]
    if names:
        case_names_ids = set(i for i, n in enumerate(case_names) if n in names)
        case_names = [case_names[i] for i in sorted(case_names_ids)]
    else:
        case_names_ids = set(range(len(case_names)))
    x = dok_matrix((len(case_names_ids), len(descr_names)), dtype=np.float32)
    with open(fname) as f:
        row = 0
        for i, line in enumerate(f):
            if i in case_names_ids:
                tmp = line.strip().split(' ')
                for v in tmp:
                    col, value = v.split(':')
                    x[row, int(col)] = value
                row += 1
    return descr_names, case_names, x.toarray()


def load_y(fname):
    with open(fname) as f:
        d = dict()
        f.readline()
        for line in f:
            tmp = line.strip().split('\t')
            d[tmp[0]] = float(tmp[1])
    return d


def save_object(model, fname):
    joblib.dump(model, fname)


def mean(l):
    return sum(l) / len(l)


def get_sensitivity(obs, pred, classes=(0, 1)):
    tp = sum(np.logical_and((pred == classes[1]), (obs == pred)))
    # tp = sum(obs + pred == 2 * classes[1])
    t = sum(obs == classes[1])
    return tp / t


def get_specificity(obs, pred, classes=(0, 1)):
    tn = sum(np.logical_and((pred == classes[0]), (obs == pred)))
    # tn = sum(obs + pred == 2 * classes[0])
    n = sum(obs == classes[0])
    return tn / n


def get_kappa(obs, pred, classes=(0, 1)):
    obs_0 = sum(obs == classes[0])
    pred_0 = sum(pred == classes[0])
    obs_1 = sum(obs == classes[1])
    pred_1 = sum(pred == classes[1])
    baseline = (obs_0 * pred_0 + obs_1 * pred_1) / (len(obs) ** 2)
    acc = sum(obs == pred) / len(obs)
    return (acc - baseline) / (1 - baseline)


def get_r2_test(obs, pred, mean_training):
    return 1 - (sum((obs - pred) ** 2) / sum((obs - mean_training) ** 2))


def get_mse(obs, pred):
    return sum((obs - pred) ** 2) / (len(pred) - 1)


def calc_model_stat_2(y, pred, model_type):

    if model_type == 'reg':

        d = {"r2": get_r2_test(y, pred, mean(y)),
             "mse": get_mse(y, pred),
             "rmse": get_mse(y, pred) ** 0.5}

    elif model_type == 'class':

        d = {"balanced accuracy": (get_sensitivity(y, pred) + get_specificity(y, pred)) / 2,
             "sensitivity": get_sensitivity(y, pred),
             "specificity": get_specificity(y, pred),
             "kappa": get_kappa(y, pred)}

    return d


def save_model_stat_2(model_name, file_name, model_params, y, pred, model_type, verbose):

    d = calc_model_stat_2(y, pred, model_type)

    if os.path.isfile(file_name):
        lines = open(file_name).readlines()
        lines = lines[0:2] + [line for line in lines[2:] if line.strip().split('\t')[1] != model_name]

    else:
        if model_type == 'reg':
            lines = ["Regression\n", "Time\tModel\tR2\tRMSE\tMSE\tOptimal_parameters\n"]
        elif model_type == 'class':
            lines = ["Classification\n", "Time\tModel\tBalanced accuracy\tSensitivity\tSpecificity\tKappa\tOptimal_parameters\n"]

    if model_type == 'reg':
        lines.append(strftime("%Y-%m-%d %H:%M:%S") + "\t" + model_name + "\t" + "{0:.2f}".format(d["r2"]) + "\t" +
             "{0:.2f}".format(d["rmse"]) + "\t" + "{0:.2f}".format(d["mse"]) + "\t" + model_params + "\n")

    elif model_type == 'class':
        lines.append(strftime("%Y-%m-%d %H:%M:%S") + "\t" + model_name + "\t" + "{0:.2f}".format(d["balanced accuracy"]) + "\t" +
                     "{0:.2f}".format(d["sensitivity"]) + "\t" + "{0:.2f}".format(d["specificity"]) + "\t" +
                     "{0:.2f}".format(d["kappa"]) + "\t" + model_params + "\n")

    if verbose:
        print(lines[-1].strip())

    open(file_name, "wt").writelines(lines)


def save_bound_box_constrains(x, fname):
    d = dict()
    d['min_values'] = tuple(np.amin(x, axis=0))
    d['max_values'] = tuple(np.amax(x, axis=0))
    save_object(d, fname)


def add_obj_to_file(fname, obj):
    with open(fname, "ab") as f:
        joblib.dump(obj, f)


def make_subsets(y, seed=42):

    np.random.seed(seed)
    ids_min = []
    subsets = []
    if len(np.where(y == 0)[0]) / len(np.where(y == 1)[0]) > 1.5:
        ids_min = list(np.where(y == 1)[0])
        ids_maj = list(np.where(y == 0)[0])
    elif len(np.where(y == 1)[0]) / len(np.where(y == 0)[0]) > 1.5:
        ids_min = list(np.where(y == 0)[0])
        ids_maj = list(np.where(y == 1)[0])

    if ids_min:
        ratio = round(len(ids_maj) / len(ids_min))
        np.random.shuffle(ids_maj)
        np.random.shuffle(ids_min)
        subset_sizes = (len(ids_maj) // ratio) * np.ones(ratio, dtype=np.int)
        subset_sizes[:len(ids_maj) % ratio] += 1
        current = 0
        for subset_size in subset_sizes:
            start, stop = current, current + subset_size
            subsets.append(ids_maj[start:stop] + ids_min)
            current = stop

    return subsets


def main_params(x_fname, y_fname, model_names, models_dir, ncores, model_type, verbose, cv_predictions, imbalanced, input_format):

    seed = 42

    # create models subdir
    if models_dir is None:
        models_dir = os.path.join(os.path.dirname(x_fname), "models")
    if not os.path.exists(models_dir):
        os.makedirs(models_dir)
    for m in model_names:
        fpath = os.path.join(models_dir, m + ".pkl")
        if os.path.exists(fpath):
            os.remove(fpath)

    model_stat_fname = os.path.join(models_dir, "models_stat.txt")

    # load y
    y = load_y(y_fname)

    # load x
    if input_format == 'txt':
        descr_names, mol_names, x = load_sirms_txt(x_fname, names=y.keys())
    elif input_format == 'svm':
        descr_names, mol_names, x = load_sirms_svm(x_fname, names=y.keys())
    else:
        print("Illegal value of input format: " % input_format)
        exit()

    # process y
    y = np.asarray([y[n] for n in mol_names])

    # process x
    save_bound_box_constrains(x, os.path.join(models_dir, "bound_box.pkl"))
    save_object(descr_names, os.path.join(models_dir, "var_names.pkl"))

    # scale
    scale = StandardScaler().fit(x)
    save_object(scale, os.path.join(models_dir, "scale.pkl"))

    x = scale.transform(x)

    if model_type == "class":
        cv = ms.StratifiedKFold(n_splits=5, random_state=seed, shuffle=True)
    elif model_type == "reg":
        cv = ms.KFold(n_splits=5, random_state=seed, shuffle=True)

    if model_type == "class" and imbalanced:
        subsets = make_subsets(y, seed)
        if not subsets:
            warnings.warn("The data set is balanced (ratio majority:minority < 1.5)."
                          "No multiple undersampling will be done", Warning)
            subsets = [list(range(y.shape[0]))]
    else:
        subsets = [list(range(y.shape[0]))]

    # build models

    for current_model in model_names:

        if verbose:
            print(current_model.upper() + ' model building...')

        models_lst = []  # this lst refreshes on each model name; here we store either 1 model in balanced case or list of models=number of subsets in the case of imbalanced

        for subset in subsets:

            if current_model == "rf":

                # choosing optimal parameters
                param_grid = {"max_features": [x.shape[1] // 10, x.shape[1] // 7, x.shape[1] // 5, x.shape[1] // 3],
                              "n_estimators": [500]}

                if model_type == "reg":
                    m = ms.GridSearchCV(RandomForestRegressor(random_state=seed), param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                elif model_type == "class":
                    m = ms.GridSearchCV(RandomForestClassifier(random_state=seed), param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                m.fit(x[subset], y[subset])

                # final model
                if model_type == "reg":
                    m = RandomForestRegressor(n_estimators=m.best_params_["n_estimators"],
                                              max_features=m.best_params_["max_features"],
                                              bootstrap=True, random_state=seed)
                elif model_type == "class":
                    m = RandomForestClassifier(n_estimators=m.best_params_["n_estimators"],
                                               max_features=m.best_params_["max_features"],
                                               bootstrap=True, random_state=seed)
                models_lst.append(m)

            if current_model == "gbm":

                # choosing optimal parameters
                param_grid = {"n_estimators": [100, 200, 300, 400, 500]}

                if model_type == "reg":
                    m = ms.GridSearchCV(GradientBoostingRegressor(subsample=0.5, max_features=0.5, random_state=seed),
                                        param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                elif model_type == "class":
                    m = ms.GridSearchCV(GradientBoostingClassifier(subsample=0.5, max_features=0.5, random_state=seed),
                                        param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                m.fit(x[subset], y[subset])

                # final model
                if model_type == "reg":
                    m = GradientBoostingRegressor(n_estimators=m.best_params_["n_estimators"],
                                                  subsample=0.5, max_features=0.5, random_state=seed)
                elif model_type == "class":
                    m = GradientBoostingClassifier(n_estimators=m.best_params_["n_estimators"],
                                                   subsample=0.5, max_features=0.5, random_state=seed)
                models_lst.append(m)

            if current_model == "svm":

                # choosing optimal parameters
                if model_type == "reg":
                    param_grid = {"C": [10 ** i for i in range(0, 5)],
                                  "epsilon": [0.35, 0.3, 0.25, 0.2, 0.15, 0.1, 0.05, 0.01]}
                    m = ms.GridSearchCV(svm.SVR(kernel='rbf'), param_grid, n_jobs=ncores, cv=cv, refit=False,
                                        verbose=verbose)
                elif model_type == "class":
                    param_grid = {"C": [10 ** i for i in range(0, 5)],
                                  "gamma": [10 ** i for i in range(-6, 0)]}
                    m = ms.GridSearchCV(svm.SVC(kernel='rbf', random_state=seed), param_grid, n_jobs=ncores, cv=cv, refit=False,
                                        verbose=verbose)
                m.fit(x[subset], y[subset])

                # final model
                if model_type == "reg":
                    m = svm.SVR(kernel='rbf', C=m.best_params_["C"], epsilon=m.best_params_["epsilon"])
                elif model_type == "class":
                    m = svm.SVC(kernel='rbf', C=m.best_params_["C"], gamma=m.best_params_["gamma"],
                                probability=True, random_state=seed)
                models_lst.append(m)

            if current_model == "pls" and model_type == "reg":

                # choosing optimal parameters
                param_grid = {"n_components": [i for i in range(1, 8)]}
                m = ms.GridSearchCV(PLSRegression(), param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                m.fit(x[subset], y[subset])

                # final model
                m = PLSRegression(n_components=m.best_params_["n_components"])
                models_lst.append(m)

            if current_model == "knn":

                # choosing optimal parameters
                param_grid = {"n_neighbors": [i for i in range(3, 21)]}
                if model_type == "reg":
                    m = ms.GridSearchCV(KNeighborsRegressor(), param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                elif model_type == "class":
                    m = ms.GridSearchCV(KNeighborsClassifier(), param_grid, n_jobs=ncores, cv=cv, refit=False, verbose=verbose)
                m.fit(x[subset], y[subset])

                # final model
                if model_type == "reg":
                    m = KNeighborsRegressor(n_neighbors=m.best_params_["n_neighbors"])
                elif model_type == "class":
                    m = KNeighborsClassifier(n_neighbors=m.best_params_["n_neighbors"])
                models_lst.append(m)

        # return cv predictions
        ncol = len(models_lst) + 1 if len(models_lst) > 1 else len(models_lst)   # +1 column if consensus
        cv_pred = np.column_stack((y, np.full((y.shape[0], ncol), np.nan)))

        for i, (m, subset) in enumerate(zip(models_lst, subsets)):
            pred = ms.cross_val_predict(estimator=m, X=x[subset], y=y[subset], cv=cv)
            if current_model == 'pls':   # reshape for pls because it returns 2d array and we need 1d
                pred = pred.reshape(len(subset))
            cv_pred[subset, i + 1] = pred

            # build final model, save it and its stat
            m.fit(x[subset], y[subset])
            add_obj_to_file(os.path.join(models_dir, current_model + '.pkl'), m)
            save_model_stat_2(current_model + '_%i' % i, model_stat_fname, str(m.get_params())[1:-1],
                              y[subset],
                              cv_pred[subset, i + 1],
                              model_type,
                              verbose)

        # calc cv consensus and save stat
        if model_type == "class" and len(models_lst) > 1:
            cv_pred[:, -1] = np.apply_along_axis(get_major_vote, 1, cv_pred[:, 1:])
            # cv_pred[:, -1] = np.around(np.nanmean(cv_pred[:, 1:], axis=1))
            save_model_stat_2(current_model + "_consensus", model_stat_fname, "",
                              y,
                              cv_pred[:, -1],
                              model_type,
                              verbose)

        # save cv predictions
        if cv_predictions:
            np.savetxt(os.path.join(models_dir, current_model + "_cv_pred.txt"),
                       np.column_stack([mol_names, np.round(cv_pred, 3)]),
                       fmt="%s",
                       delimiter="\t",
                       comments="",
                       header="Mol\tObs\t" +
                              "\t".join("%s_%i" % (current_model, i) for i in range(len(models_lst))) +
                              "\t" + current_model + "_consensus")

        if verbose:
            print(current_model.upper() + ' model was built\n')


def main():

    parser = argparse.ArgumentParser(description='Build QSAR/QSPR models.')
    parser.add_argument('-x', '--descriptors', metavar='descriptors.txt', required=True,
                        help='input file with descriptors in txt or svm formats.')
    parser.add_argument('-f', '--descriptors_format', metavar='txt', default='txt',
                        help='format of the input file with descriptors (txt|svm). Svm file has to have two '
                             'supplementary files with the same name and extensions rownames and colnames. '
                             'Default txt.')
    parser.add_argument('-y', '--property', metavar='property.txt', required=True,
                        help='input text file with property/activity values.')
    parser.add_argument('-m', '--models', metavar='[rf gbm svm pls knn]', default=['rf'], nargs='*',
                        help='models to build.')
    parser.add_argument('-d', '--models_dir', default=None, help='path to dir where models will be saved.')
    parser.add_argument('-c', '--ncores', metavar='1|2|3|...|all', default=str(cpu_count() - 1),
                        help='number of cores used for models building. Default: all cores - 1.')
    parser.add_argument('-t', '--model_type', metavar='reg|class', default='reg',
                        help='models type: reg for regression and class for classification.')
    parser.add_argument('-v', '--verbose', default=1,
                        help='Integer value. 0 - print no details. 1 and more - verbose output. Default: 1.')
    parser.add_argument('--imbalanced', default=False, action='store_true',
                        help='if set this flag the input data set will be considered imbalanced and '
                             'multiple undersampling will be applied. Works only with classification tasks '
                             'when the ratio between two classes is >= 1.5.')
    parser.add_argument('-p', '--cv_predictions', action='store_true', default=True,
                        help='True/False to save cross-validation predictions in text files along '
                             'with model files.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "descriptors": x_fname = v
        if o == "property": y_fname = v
        if o == "models": model_names = v
        if o == "ncores":
            if v == 'all':
                ncores = cpu_count()
            else:
                ncores = int(v)
                if ncores > cpu_count():
                    ncores = cpu_count()
        if o == "model_type": model_type = v
        if o == "verbose": verbose = int(v)
        if o == "cv_predictions": cv_predictions = v
        if o == "descriptors_format": input_format = v
        if o == "models_dir": models_dir = v
        if o == "imbalanced": imbalanced = v
    if input_format not in ['txt', 'svm']:
        print("Input file format is wrong - %s. Only txt and svm are allowed." % input_format)
        exit()

    main_params(x_fname=x_fname,
                y_fname=y_fname,
                model_names=model_names,
                models_dir=models_dir,
                ncores=ncores,
                model_type=model_type,
                verbose=verbose,
                cv_predictions=cv_predictions,
                imbalanced=imbalanced,
                input_format=input_format)


if __name__ == '__main__':
    main()
